import re
import json

from urllib.parse import parse_qs, urlparse

from json import JSONDecodeError

from django.template.defaultfilters import truncatechars
from django.utils.encoding import force_text

from security.config import settings
from security.utils import remove_nul_from_string


def is_base_collection(v):
    return isinstance(v, (list, tuple, set))


def get_headers(request):
    regex = re.compile('^HTTP_')
    return dict((regex.sub('', header), value) for (header, value)
                in request.META.items() if header.startswith('HTTP_'))


def regex_sub_groups_global(pattern, repl, string):
    """
    Globally replace all groups inside pattern with `repl`.
    If `pattern` doesn't have groups the whole match is replaced.
    """
    for search in reversed(list(re.finditer(pattern, string))):
        for i in range(len(search.groups()), 0 if search.groups() else -1, -1):
            start, end = search.span(i)
            string = string[:start] + repl + string[end:]
    return string


def flat_params(params):
    return {
        k: v[0] if is_base_collection(v) and len(v) == 1 else v
        for k, v in params.items()
    }


def list_params(params):
    return {
        k: list(v) if is_base_collection(v) else [v]
        for k, v in params.items()
    }


def get_logged_params(url):
    return flat_params(parse_qs(urlparse(url).query))


def hide_sensitive_data_body(content):
    if settings.HIDE_SENSITIVE_DATA:
        for pattern in settings.HIDE_SENSITIVE_DATA_PATTERNS.get('BODY', ()):
            content = regex_sub_groups_global(pattern, settings.SENSITIVE_DATA_REPLACEMENT, content)
    return content


def hide_sensitive_data_headers(headers):
    if settings.HIDE_SENSITIVE_DATA:
        headers = dict(headers)
        for pattern in settings.HIDE_SENSITIVE_DATA_PATTERNS.get('HEADERS', ()):
            for header_name, header in headers.items():
                if re.match(pattern, header_name, re.IGNORECASE):
                    headers[header_name] = settings.SENSITIVE_DATA_REPLACEMENT
    return headers


def hide_sensitive_data_queries(queries):
    if settings.HIDE_SENSITIVE_DATA:
        queries = dict(queries)
        for pattern in settings.HIDE_SENSITIVE_DATA_PATTERNS.get('QUERIES', ()):
            for query_name, query in queries.items():
                if re.match(pattern, query_name, re.IGNORECASE):
                    queries[query_name] = (
                        len(query) * [settings.SENSITIVE_DATA_REPLACEMENT] if is_base_collection(query)
                        else settings.SENSITIVE_DATA_REPLACEMENT
                    )
    return queries


def truncate_json_data(data):
    if isinstance(data, dict):
        return {key: truncate_json_data(val) for key, val in data.items()}
    elif isinstance(data, list):
        return [truncate_json_data(val) for val in data]
    elif isinstance(data, str):
        return truncatechars(data, settings.LOG_JSON_STRING_LENGTH)
    else:
        return data


def truncate_body(content, max_length):
    content = force_text(content, errors='replace')
    if len(content) > max_length:
        try:
            json_content = json.loads(content)
            return (
                json.dumps(truncate_json_data(json_content))
                if isinstance(json_content, (dict, list)) and settings.LOG_JSON_STRING_LENGTH is not None
                else content[:max_length + 1]
            )
        except JSONDecodeError:
            return content[:max_length + 1]
    else:
        return content


def clean_body(body, max_length):
    if body is None:
        return body
    body = force_text(body, errors='replace')
    cleaned_body = truncatechars(
        truncate_body(body, max_length), max_length + len(settings.SENSITIVE_DATA_REPLACEMENT)
    ) if max_length is not None else str(body)
    cleaned_body = hide_sensitive_data_body(remove_nul_from_string(cleaned_body)) if cleaned_body else cleaned_body
    cleaned_body = truncatechars(cleaned_body, max_length) if max_length else cleaned_body
    return cleaned_body


def clean_json(data):
    return {remove_nul_from_string(k): remove_nul_from_string(v) if isinstance(v, str) else v for k, v in data.items()}


def clean_headers(headers):
    return hide_sensitive_data_headers(clean_json(headers)) if headers else headers


def clean_queries(queries):
    return hide_sensitive_data_queries(clean_json(queries)) if queries else queries


def log_input_request_with_data(request, related_objects=None, slug=None, extra_data=None):
    input_request_logger = getattr(request, 'input_request_logger', None)
    if not input_request_logger:
        return False
    if related_objects:
        input_request_logger.add_related_objects(*related_objects)
    if slug:
        input_request_logger.set_slug(slug)
    if extra_data:
        input_request_logger.update_extra_data(extra_data)
    return True



from django.db.models.signals import pre_migrate, post_migrate

from security.backends.app import SecurityBackend

from . import db_helpers


def start_migration(sender, **kwargs):
    db_helpers.is_running_migration = True


def end_migration(sender, **kwargs):
    db_helpers.is_running_migration = False


class SecuritySQLBackend(SecurityBackend):

    name = 'security.backends.sql'
    label = 'security_backends_sql'
    backend_name = 'sql'
    writer = 'security.backends.sql.writer.SQLBackendWriter'
    reader = 'security.backends.sql.reader.SQLBackendReader'

    def ready(self):
        super().ready()
        pre_migrate.connect(start_migration, sender=self)
        post_migrate.connect(end_migration, sender=self)



from django.utils.timezone import now

from security.enums import LoggerName
from security.logging.common import SecurityLogger
from security.backends.signals import (
    command_started, command_output_updated, command_finished, command_error,
)


class CommandLogger(SecurityLogger):

    name = LoggerName.COMMAND

    def log_started(self, name, input, is_executed_from_command_line, parent_log=None):
        self.data.update(dict(
            name=name,
            input=input,
            is_executed_from_command_line=is_executed_from_command_line,
            start=now()
        ))
        command_started.send(sender=CommandLogger, logger=self)

    def log_finished(self):
        self.data.update(dict(
            stop=now()
        ))
        command_finished.send(sender=CommandLogger, logger=self)

    def log_output_updated(self, output):
        self.data['output'] = output
        command_output_updated.send(sender=CommandLogger, logger=self)

    def log_exception(self, ex_tb):
        self.data.update(dict(
            error_message=ex_tb,
            stop=now()
        ))
        command_error.send(sender=CommandLogger, logger=self)



from datetime import timedelta

from django.core.management.base import CommandError
from django.core.management.base import BaseCommand
from django.utils.timezone import now

from security.enums import LoggerName
from security.backends.writer import clean_logs


UNIT_OPTIONS = {
    'h': lambda amount: now() - timedelta(hours=amount),
    'd': lambda amount: now() - timedelta(days=amount),
    'w': lambda amount: now() - timedelta(weeks=amount),
    'm': lambda amount: now() - timedelta(days=(30 * amount)),  # 30-day month
    'y': lambda amount: now() - timedelta(weeks=(52 * amount)),  # 364-day year
}


class Command(BaseCommand):

    def add_arguments(self, parser):
        parser.add_argument('--type', action='store', dest='type', default='input-request',
                            help='Tells Django what type of requests should be removed '
                                 '({}).'.format('/'.join([logger.value for logger in LoggerName])))
        parser.add_argument('--expiration', action='store', dest='expiration',
                            help='Sets the timedelta from which logs will be removed.', required=True)
        parser.add_argument('--noinput', action='store_false', dest='interactive', default=True,
                            help='Tells Django to NOT prompt the user for input of any kind.')
        parser.add_argument('--backup', action='store', dest='backup', default=False,
                            help='Tells Django where to backup removed logs.')

    def handle(self, expiration, type, **options):
        # Check we have the correct values
        unit = expiration[-1]
        amount = expiration[0:-1]
        try:
            amount = int(amount)
        except ValueError:
            raise CommandError('Invalid expiration format')

        if unit not in UNIT_OPTIONS:
            raise CommandError('Invalid expiration format')

        try:
            type = LoggerName(type)
        except KeyError:
            raise CommandError('Type can be only one of "{}"'.format('/'.join([logger.value for logger in LoggerName])))

        if options.get('interactive'):
            confirm = input('''
               You have requested a database reset.
               This will IRREVERSIBLY DESTROY any
               logs created before {} {}
               ago. Are you sure you want to do this?
               Type 'yes' to continue, or 'no' to cancel: '''.format(amount, unit))
        else:
            confirm = 'yes'

        if confirm == 'yes':
            self.stdout.write('Clean data')
            clean_logs(type, UNIT_OPTIONS[unit](amount), options.get('backup'), self.stdout)



import json
import responses

from requests.exceptions import ConnectionError

from django.test import override_settings

from germanium.decorators import data_consumer
from germanium.test_cases.client import ClientTestCase
from germanium.tools import (
    assert_equal, assert_raises, assert_not_in, assert_in, assert_equal_model_fields, assert_is_not_none,
    assert_length_equal, all_eq_obj
)

from security import requests
from security.backends.signals import (
    output_request_started, output_request_finished, output_request_error
)
from security.decorators import log_with_data
from security.enums import RequestLogState
from security.backends.sql.models import OutputRequestLog as SQLOutputRequestLog
from security.backends.elasticsearch.models import OutputRequestLog as ElasticsearchOutputRequestLog
from security.backends.testing import capture_security_logs

from .base import BaseTestCaseMixin, TRUNCATION_CHAR


@override_settings(SECURITY_BACKEND_WRITERS={})
class OutputRequestLogTestCase(BaseTestCaseMixin, ClientTestCase):

    @responses.activate
    @data_consumer('create_user')
    def test_output_request_should_be_logged(self, user):
        responses.add(responses.POST, 'https://localhost/test', body='test')
        expected_output_request_started_data = {
            'request_headers': {
                'User-Agent': 'python-requests/2.26.0',
                'Accept-Encoding': 'gzip, deflate',
                'Accept': '*/*',
                'Connection': 'keep-alive',
                'Content-Length': '16'
            },
            'request_body': '{"test": "test"}',
            'method': 'POST',
            'host': 'localhost',
            'path': '/test',
            'queries': {},
            'is_secure': True,
            'start': all_eq_obj,
        }
        expected_output_request_finished_data = {
            **expected_output_request_started_data,
            'stop': all_eq_obj,
            'response_code': 200,
            'response_headers': {'Content-Type': 'text/plain'},
            'response_body': 'test',
        }
        with capture_security_logs() as logged_data:
            requests.post(
                'https://localhost/test',
                data=json.dumps({'test': 'test'}),
                slug='test',
                related_objects=[user]
            )
            assert_length_equal(logged_data.output_request_started, 1)
            assert_length_equal(logged_data.output_request_finished, 1)
            assert_length_equal(logged_data.output_request_error, 0)
            assert_equal(logged_data.output_request_started[0].data, expected_output_request_started_data)
            assert_equal(logged_data.output_request_finished[0].data, expected_output_request_finished_data)
            assert_equal(logged_data.output_request_started[0].slug, 'test')
            assert_equal(logged_data.output_request_finished[0].related_objects, [user])

    @responses.activate
    def test_output_request_error_should_be_logged(self):
        expected_output_request_started_data = {
            'request_headers': {
                'User-Agent': 'python-requests/2.26.0',
                'Accept-Encoding': 'gzip, deflate',
                'Accept': '*/*',
                'Connection': 'keep-alive',
                'Content-Length': '16'
            },
            'request_body': '{"test": "test"}',
            'method': 'POST',
            'host': 'localhost',
            'path': '/test',
            'queries': {},
            'is_secure': True,
            'start': all_eq_obj,
        }
        expected_output_request_error_data = {
            **expected_output_request_started_data,
            'stop': all_eq_obj,
            'error_message': all_eq_obj
        }
        with capture_security_logs() as logged_data:
            with assert_raises(ConnectionError):
                requests.post(
                    'https://localhost/test',
                    data=json.dumps({'test': 'test'}),
                )
            assert_length_equal(logged_data.output_request_started, 1)
            assert_length_equal(logged_data.output_request_finished, 0)
            assert_length_equal(logged_data.output_request_error, 1)
            assert_equal(logged_data.output_request_started[0].data, expected_output_request_started_data)
            assert_equal(logged_data.output_request_error[0].data, expected_output_request_error_data)

    @responses.activate
    def test_response_sensitive_data_body_in_json_should_be_hidden(self):
        responses.add(responses.POST, 'http://localhost', body='test')

        with capture_security_logs() as logged_data:
            requests.post('http://localhost', data=json.dumps({'password': 'secret-password'}))
            assert_in('"password": "[Filtered]"', logged_data.output_request[0].data['request_body'])
            assert_not_in('"password": "secret-password"', logged_data.output_request[0].data['request_body'])
            assert_in('"password": "secret-password"', responses.calls[0].request.body)
            assert_not_in('"password": "[Filtered]"', responses.calls[0].request.body)

    @responses.activate
    def test_response_sensitive_headers_should_be_hidden(self):
        responses.add(responses.POST, 'http://localhost', body='test')
        with capture_security_logs() as logged_data:
            requests.post('http://localhost', headers={'token': 'secret'})
            assert_equal(logged_data.output_request[0].data['request_headers']['token'], '[Filtered]')
            assert_equal(responses.calls[0].request.headers['token'], 'secret')

    @responses.activate
    def test_response_sensitive_params_data_should_be_hidden(self):
        responses.add(responses.POST, 'http://localhost', body='test')
        with capture_security_logs() as logged_data:
            requests.post('http://localhost', params={'token': 'secret'})
            assert_equal(logged_data.output_request[0].data['queries']['token'], '[Filtered]')
            assert_equal(responses.calls[0].request.url, 'http://localhost/?token=secret')

    @responses.activate
    def test_response_more_sensitive_params_data_should_be_hidden(self):
        responses.add(responses.POST, 'http://localhost', body='test')
        with capture_security_logs() as logged_data:
            requests.post('http://localhost', params={'token': ['secret', 'secret2']})
            assert_equal(logged_data.output_request[0].data['queries']['token'], ['[Filtered]', '[Filtered]'])
            assert_equal(responses.calls[0].request.url, 'http://localhost/?token=secret&token=secret2')

    @responses.activate
    def test_response_sensitive_params_and_url_query_together_data_should_be_logged(self):
        responses.add(responses.POST, 'http://localhost', body='test')
        with capture_security_logs() as logged_data:
            requests.post('http://localhost?a=1&a=2', params={'b': '6', 'a': '3', 'c': ['5']})
            assert_equal(logged_data.output_request[0].data['queries'], {'b': '6', 'a': ['1', '2', '3'], 'c': '5'})

    @responses.activate
    @override_settings(SECURITY_BACKEND_WRITERS={'sql'})
    @data_consumer('create_user')
    def test_output_request_should_be_logged_in_sql_backend(self, user):
        responses.add(responses.POST, 'https://localhost/test', body='test')
        requests.post(
            'https://localhost/test',
            data=json.dumps({'test': 'test'}),
            slug='test',
            related_objects=[user]
        )
        assert_equal(SQLOutputRequestLog.objects.count(), 1)
        sql_output_request_log = SQLOutputRequestLog.objects.get()
        assert_equal_model_fields(
            sql_output_request_log,
            request_headers={
                'User-Agent': 'python-requests/2.26.0',
                'Accept-Encoding': 'gzip, deflate',
                'Accept': '*/*',
                'Connection': 'keep-alive',
                'Content-Length': '16'
            },
            request_body='{"test": "test"}',
            method='POST',
            host='localhost',
            path='/test',
            queries={},
            is_secure=True,
            slug='test',
            time=(sql_output_request_log.stop - sql_output_request_log.start).total_seconds(),
            extra_data={},
            error_message=None,
            response_code=200,
            response_headers={'Content-Type': 'text/plain'},
            response_body='test',
            state=RequestLogState.INFO,
        )
        assert_equal([rel_obj.object for rel_obj in sql_output_request_log.related_objects.all()], [user])

    @responses.activate
    @override_settings(SECURITY_BACKEND_WRITERS={'elasticsearch'})
    @data_consumer('create_user')
    def test_output_request_should_be_logged_in_elasticsearch_backend(self, user):
        responses.add(responses.POST, 'https://localhost/test', body='test')
        with capture_security_logs() as logged_data:
            requests.post(
                'https://localhost/test',
                data=json.dumps({'test': 'test'}),
                slug='test',
                related_objects=[user]
            )
            elasticsearch_output_request_log = ElasticsearchOutputRequestLog.get(
                id=logged_data.output_request[0].id
            )
            assert_equal_model_fields(
                elasticsearch_output_request_log,
                request_headers='{"User-Agent": "python-requests/2.26.0", "Accept-Encoding": "gzip, deflate", '
                                '"Accept": "*/*", "Connection": "keep-alive", "Content-Length": "16"}',
                request_body='{"test": "test"}',
                method='POST',
                host='localhost',
                path='/test',
                queries='{}',
                is_secure=True,
                slug='test',
                time=(elasticsearch_output_request_log.stop - elasticsearch_output_request_log.start).total_seconds(),
                extra_data={},
                error_message=None,
                response_code=200,
                response_headers='{"Content-Type": "text/plain"}',
                response_body='test',
                state=RequestLogState.INFO,
            )
            assert_equal(
                [rel_obj for rel_obj in elasticsearch_output_request_log.related_objects],
                ['default|3|{}'.format(user.id)]
            )

    @responses.activate
    @override_settings(SECURITY_BACKEND_WRITERS={'logging'})
    @data_consumer('create_user')
    def test_output_request_should_be_logged_in_logging_backend(self, user):
        responses.add(responses.POST, 'https://localhost/test', body='test')
        with capture_security_logs() as logged_data:
            with self.assertLogs('security.output_request', level='INFO') as cm:
                requests.post(
                    'https://localhost/test',
                    data=json.dumps({'test': 'test'}),
                    slug='test',
                    related_objects=[user]
                )
                assert_equal(
                    cm.output,
                    [
                        f'INFO:security.output_request:'
                        f'Output request "{logged_data.output_request[0].id}" '
                        f'to "localhost" with path "/test" was started',
                        f'INFO:security.output_request:'
                        f'Output request "{logged_data.output_request[0].id}" '
                        f'to "localhost" with path "/test" was successful'
                    ]
                )

    @responses.activate
    @override_settings(SECURITY_BACKEND_WRITERS={'sql'})
    def test_error_output_request_should_be_logged_in_sql_backend(self,):
        with assert_raises(ConnectionError):
            requests.post(
                'https://localhost/test',
                data=json.dumps({'test': 'test'}),
                slug='test',
            )
        assert_equal(SQLOutputRequestLog.objects.count(), 1)
        sql_output_request_log = SQLOutputRequestLog.objects.get()
        assert_equal_model_fields(
            sql_output_request_log,
            request_headers={
                'User-Agent': 'python-requests/2.26.0',
                'Accept-Encoding': 'gzip, deflate',
                'Accept': '*/*',
                'Connection': 'keep-alive',
                'Content-Length': '16'
            },
            request_body='{"test": "test"}',
            method='POST',
            host='localhost',
            path='/test',
            queries={},
            is_secure=True,
            slug='test',
            time=(sql_output_request_log.stop - sql_output_request_log.start).total_seconds(),
            extra_data={},
            response_code=None,
            response_headers=None,
            response_body=None,
            state=RequestLogState.ERROR,
        )
        assert_is_not_none(sql_output_request_log.error_message)

    @responses.activate
    @override_settings(SECURITY_BACKEND_WRITERS={'elasticsearch'})
    def test_error_output_request_should_be_logged_in_elasticsearch_backend(self):
        with capture_security_logs() as logged_data:
            with assert_raises(ConnectionError):
                requests.post(
                    'https://localhost/test',
                    data=json.dumps({'test': 'test'}),
                    slug='test',
                )
            elasticsearch_input_request_log = ElasticsearchOutputRequestLog.get(
                id=logged_data.output_request[0].id
            )
            assert_equal_model_fields(
                elasticsearch_input_request_log,
                request_headers='{"User-Agent": "python-requests/2.26.0", "Accept-Encoding": "gzip, deflate", '
                                '"Accept": "*/*", "Connection": "keep-alive", "Content-Length": "16"}',
                request_body='{"test": "test"}',
                method='POST',
                host='localhost',
                path='/test',
                queries='{}',
                is_secure=True,
                slug='test',
                time=(elasticsearch_input_request_log.stop - elasticsearch_input_request_log.start).total_seconds(),
                extra_data={},
                response_code=None,
                response_headers=None,
                response_body=None,
                state=RequestLogState.ERROR,
            )
            assert_is_not_none(elasticsearch_input_request_log.error_message)

    @responses.activate
    @override_settings(SECURITY_BACKEND_WRITERS={'logging'})
    def test_error_output_request_should_be_logged_in_logging_backend(self):
        with capture_security_logs() as logged_data:
            with self.assertLogs('security.output_request', level='INFO') as cm:
                with assert_raises(ConnectionError):
                    requests.post(
                        'https://localhost/test',
                        data=json.dumps({'test': 'test'}),
                        slug='test',
                    )
                assert_equal(
                    cm.output,
                    [
                        f'INFO:security.output_request:'
                        f'Output request "{logged_data.output_request[0].id}" '
                        f'to "localhost" with path "/test" was started',
                        f'ERROR:security.output_request:'
                        f'Output request "{logged_data.output_request[0].id}" '
                        f'to "localhost" with path "/test" failed'
                    ]
                )

    @responses.activate
    @data_consumer('create_user')
    def test_slug_and_related_data_should_be_send_to_output_request_logger(self, user):
        responses.add(responses.POST, 'https://localhost/test', body='test')
        with log_with_data(related_objects=[user], slug='TEST'):
            with capture_security_logs() as logged_data:
                requests.post(
                    'https://localhost/test',
                    data=json.dumps({'test': 'test'}),
                )
                assert_equal(logged_data.output_request[0].related_objects, {user})
                assert_equal(logged_data.output_request[0].slug, 'TEST')



import os
import json
import gzip
import math

from io import TextIOWrapper

from datetime import datetime, time, timedelta

from django.core.serializers.json import DjangoJSONEncoder
from django.utils.timezone import now, utc
from django.utils.module_loading import import_string

from celery import current_app
from celery.exceptions import NotRegistered

from elasticsearch import ConflictError
from elasticsearch_dsl import Q

from security.config import settings
from security.enums import (
    RequestLogState, CeleryTaskInvocationLogState, CeleryTaskRunLogState, CommandState
)
from security.backends.common.helpers import get_parent_log_key_or_none
from security.backends.writer import BaseBackendWriter

from .models import (
    InputRequestLog, OutputRequestLog, CommandLog, CeleryTaskRunLog, CeleryTaskInvocationLog, get_key_from_object,
    get_response_state, get_log_model_from_logger_name
)


def lazy_serialize_qs_without_pk(qs):
    def generator():
        for obj in qs:
            yield obj.to_dict()

    class StreamList(list):
        def __iter__(self):
            return generator()

        def __len__(self):
            return 1

    return StreamList()


def get_querysets_by_batch(qs, batch):
    steps = math.ceil(qs.count() / batch)
    for step in range(steps):
        yield qs[step * batch:(step+1) * batch]


class ElasticsearchBackendWriter(BaseBackendWriter):

    def _get_related_object_keys(self, logger):
        return [
            get_key_from_object(related_object) for related_object in logger.related_objects
        ]

    def input_request_started(self, logger):
        input_request_log = InputRequestLog(
            slug=logger.slug,
            release=logger.release,
            related_objects=self._get_related_object_keys(logger),
            extra_data=logger.extra_data,
            state=RequestLogState.INCOMPLETE,
            parent_log=get_parent_log_key_or_none(logger),
            **logger.data
        )
        input_request_log.meta.id = logger.id

        input_request_log.save()
        logger.backend_logs.elasticsearch = input_request_log

    def input_request_finished(self, logger):
        if 'elasticsearch' in logger.backend_logs:
            logger.backend_logs.elasticsearch.update(
                slug=logger.slug,
                related_objects=self._get_related_object_keys(logger),
                extra_data=logger.extra_data,
                state=get_response_state(logger.data['response_code']),
                refresh=settings.ELASTICSEARCH_AUTO_REFRESH,
                update_only_changed_fields=True,
                retry_on_conflict=1,
                **logger.data
            )

    def input_request_error(self, logger):
        if 'elasticsearch' in logger.backend_logs:
            logger.backend_logs.elasticsearch.update(
                slug=logger.slug,
                related_objects=self._get_related_object_keys(logger),
                extra_data=logger.extra_data,
                refresh=settings.ELASTICSEARCH_AUTO_REFRESH,
                update_only_changed_fields=True,
                retry_on_conflict=1,
                **logger.data
            )

    def output_request_started(self, logger):
        output_request_log = OutputRequestLog(
            slug=logger.slug,
            release=logger.release,
            related_objects=self._get_related_object_keys(logger),
            extra_data=logger.extra_data,
            state=RequestLogState.INCOMPLETE,
            update_only_changed_fields=True,
            parent_log=get_parent_log_key_or_none(logger),
            **logger.data
        )
        output_request_log.meta.id = logger.id
        output_request_log.save()
        logger.backend_logs.elasticsearch = output_request_log

    def output_request_finished(self, logger):
        if 'elasticsearch' in logger.backend_logs:
            logger.backend_logs.elasticsearch.update(
                slug=logger.slug,
                related_objects=self._get_related_object_keys(logger),
                extra_data=logger.extra_data,
                state=get_response_state(logger.data['response_code']),
                refresh=settings.ELASTICSEARCH_AUTO_REFRESH,
                update_only_changed_fields=True,
                retry_on_conflict=1,
                **logger.data
            )

    def output_request_error(self, logger):
        if 'elasticsearch' in logger.backend_logs:
            logger.backend_logs.elasticsearch.update(
                slug=logger.slug,
                related_objects=self._get_related_object_keys(logger),
                extra_data=logger.extra_data,
                state=RequestLogState.ERROR,
                refresh=settings.ELASTICSEARCH_AUTO_REFRESH,
                update_only_changed_fields=True,
                retry_on_conflict=1,
                **logger.data
            )

    def command_started(self, logger):
        command_log = CommandLog(
            slug=logger.slug,
            release=logger.release,
            related_objects=self._get_related_object_keys(logger),
            extra_data=logger.extra_data,
            state=CommandState.ACTIVE,
            parent_log=get_parent_log_key_or_none(logger),
            **logger.data
        )
        command_log.meta.id = logger.id
        command_log.save()
        logger.backend_logs.elasticsearch = command_log

    def command_output_updated(self, logger):
        if 'elasticsearch' in logger.backend_logs:
            logger.backend_logs.elasticsearch.update(
                slug=logger.slug,
                refresh=settings.ELASTICSEARCH_AUTO_REFRESH,
                update_only_changed_fields=True,
                retry_on_conflict=1,
                **logger.data
            )

    def command_finished(self, logger):
        if 'elasticsearch' in logger.backend_logs:
            logger.backend_logs.elasticsearch.update(
                slug=logger.slug,
                related_objects=self._get_related_object_keys(logger),
                extra_data=logger.extra_data,
                state=CommandState.SUCCEEDED,
                refresh=settings.ELASTICSEARCH_AUTO_REFRESH,
                update_only_changed_fields=True,
                retry_on_conflict=1,
                **logger.data
            )

    def command_error(self, logger):
        if 'elasticsearch' in logger.backend_logs:
            logger.backend_logs.elasticsearch.update(
                slug=logger.slug,
                related_objects=self._get_related_object_keys(logger),
                extra_data=logger.extra_data,
                state=CommandState.FAILED,
                refresh=settings.ELASTICSEARCH_AUTO_REFRESH,
                update_only_changed_fields=True,
                retry_on_conflict=1,
                **logger.data
            )

    def celery_task_invocation_started(self, logger):
        celery_task_invocation_log = CeleryTaskInvocationLog(
            slug=logger.slug,
            release=logger.release,
            related_objects=self._get_related_object_keys(logger),
            extra_data=logger.extra_data,
            state=CeleryTaskInvocationLogState.WAITING,
            parent_log=get_parent_log_key_or_none(logger),
            **logger.data
        )
        celery_task_invocation_log.meta.id = logger.id
        celery_task_invocation_log.save()
        logger.backend_logs.elasticsearch = celery_task_invocation_log

    def celery_task_invocation_triggered(self, logger):
        if 'elasticsearch' in logger.backend_logs:
            logger.backend_logs.elasticsearch.update(
                slug=logger.slug,
                extra_data=logger.extra_data,
                state=CeleryTaskInvocationLogState.TRIGGERED,
                refresh=settings.ELASTICSEARCH_AUTO_REFRESH,
                update_only_changed_fields=True,
                retry_on_conflict=1,
                **logger.data
            )

    def celery_task_invocation_ignored(self, logger):
        if 'elasticsearch' in logger.backend_logs:
            logger.backend_logs.elasticsearch.update(
                slug=logger.slug,
                extra_data=logger.extra_data,
                state=CeleryTaskInvocationLogState.IGNORED,
                refresh=settings.ELASTICSEARCH_AUTO_REFRESH,
                update_only_changed_fields=True,
                retry_on_conflict=1,
                **logger.data
            )

    def celery_task_invocation_timeout(self, logger):
        if 'elasticsearch' in logger.backend_logs:
            logger.backend_logs.elasticsearch.update(
                slug=logger.slug,
                extra_data=logger.extra_data,
                state=CeleryTaskInvocationLogState.TIMEOUT,
                refresh=settings.ELASTICSEARCH_AUTO_REFRESH,
                update_only_changed_fields=True,
                retry_on_conflict=1,
                **logger.data
            )

    def celery_task_invocation_expired(self, logger):
        celery_task_invocation_log = CeleryTaskInvocationLog.get(id=logger.id)
        celery_task_invocation_log.update(
            slug=logger.slug,
            state=CeleryTaskInvocationLogState.EXPIRED,
            refresh=settings.ELASTICSEARCH_AUTO_REFRESH,
            update_only_changed_fields=True,
            **logger.data
        )
        if celery_task_invocation_log.celery_task_id:
            CeleryTaskRunLog._index.refresh()
            celery_task_run_log_qs = CeleryTaskRunLog.search().filter(
                Q('term', celery_task_id=celery_task_invocation_log.celery_task_id)
                & Q('term', state=CeleryTaskRunLogState.ACTIVE.name)
            )
            for celery_task_run in celery_task_run_log_qs:
                celery_task_run.update(
                    state=CeleryTaskRunLogState.EXPIRED,
                    stop=logger.data['stop']
                )

    def celery_task_run_started(self, logger):
        celery_task_run_log = CeleryTaskRunLog(
            slug=logger.slug,
            release=logger.release,
            related_objects=self._get_related_object_keys(logger),
            extra_data=logger.extra_data,
            state=CeleryTaskRunLogState.ACTIVE,
            parent_log=get_parent_log_key_or_none(logger),
            **logger.data
        )
        celery_task_run_log.meta.id = logger.id
        celery_task_run_log.save()
        logger.backend_logs.elasticsearch = celery_task_run_log

    def celery_task_run_succeeded(self, logger):
        if 'elasticsearch' in logger.backend_logs:
            celery_task_run_log = logger.backend_logs.elasticsearch
            celery_task_run_log.update(
                slug=logger.slug,
                related_objects=self._get_related_object_keys(logger),
                extra_data=logger.extra_data,
                state=CeleryTaskRunLogState.SUCCEEDED,
                refresh=settings.ELASTICSEARCH_AUTO_REFRESH,
                update_only_changed_fields=True,
                retry_on_conflict=1,
                **logger.data
            )

            CeleryTaskInvocationLog._index.refresh()
            celery_task_invocations_qs = CeleryTaskInvocationLog.search().filter(
                'term', celery_task_id=celery_task_run_log.celery_task_id
            ).query(
                Q('term', state=CeleryTaskInvocationLogState.WAITING.name)
                | Q('term', state=CeleryTaskInvocationLogState.TRIGGERED.name)
                | Q('term', state=CeleryTaskInvocationLogState.ACTIVE.name)
            )
            for celery_task_invocation in celery_task_invocations_qs:
                try:
                    celery_task_invocation.update(
                        state=CeleryTaskInvocationLogState.SUCCEEDED,
                        stop=logger.data['stop'],
                        refresh=settings.ELASTICSEARCH_AUTO_REFRESH,
                        update_only_changed_fields=True
                    )
                except ConflictError:
                    # conflict errors are ignored, celery task invocation was changed with another process
                    pass

    def celery_task_run_failed(self, logger):
        if 'elasticsearch' in logger.backend_logs:
            celery_task_run_log = logger.backend_logs.elasticsearch
            celery_task_run_log.update(
                slug=logger.slug,
                related_objects=self._get_related_object_keys(logger),
                extra_data=logger.extra_data,
                state=CeleryTaskRunLogState.FAILED,
                refresh=settings.ELASTICSEARCH_AUTO_REFRESH,
                update_only_changed_fields=True,
                retry_on_conflict=1,
                **logger.data
            )

            CeleryTaskInvocationLog._index.refresh()
            celery_task_invocations_qs = CeleryTaskInvocationLog.search().filter(
                'term', celery_task_id=celery_task_run_log.celery_task_id
            ).query(
                Q('term', state=CeleryTaskInvocationLogState.WAITING.name)
                | Q('term', state=CeleryTaskInvocationLogState.TRIGGERED.name)
                | Q('term', state=CeleryTaskInvocationLogState.ACTIVE.name)
            )
            for celery_task_invocation in celery_task_invocations_qs:
                try:
                    celery_task_invocation.update(
                        state=CeleryTaskInvocationLogState.FAILED,
                        refresh=settings.ELASTICSEARCH_AUTO_REFRESH,
                        stop=logger.data['stop'],
                        update_only_changed_fields=True
                    )
                except ConflictError:
                    # conflict errors are ignored, celery task invocation was changed with another process
                    pass

    def celery_task_run_retried(self, logger):
        if 'elasticsearch' in logger.backend_logs:
            logger.backend_logs.elasticsearch.update(
                slug=logger.slug,
                related_objects=self._get_related_object_keys(logger),
                extra_data=logger.extra_data,
                state=CeleryTaskRunLogState.RETRIED,
                refresh=settings.ELASTICSEARCH_AUTO_REFRESH,
                update_only_changed_fields=True,
                retry_on_conflict=1,
                **logger.data
            )

    def celery_task_run_output_updated(self, logger):
        if 'elasticsearch' in logger.backend_logs:
            logger.backend_logs.elasticsearch.update(
                slug=logger.slug,
                refresh=settings.ELASTICSEARCH_AUTO_REFRESH,
                update_only_changed_fields=True,
                retry_on_conflict=1,
                **logger.data,
            )

    def set_stale_celery_task_log_state(self):
        processsing_stale_tasks = CeleryTaskInvocationLog.search().filter(
            Q('range', stale_at={'lt': now()}) & Q(
                Q('term', state=CeleryTaskInvocationLogState.WAITING.name)
                | Q('term', state=CeleryTaskInvocationLogState.TRIGGERED.name)
                | Q('term', state=CeleryTaskInvocationLogState.ACTIVE.name)
            )
        ).sort('stale_at')
        for task in processsing_stale_tasks[:settings.SET_STALE_CELERY_INVOCATIONS_LIMIT_PER_RUN]:
            task_last_run = task.last_run
            if task_last_run and task_last_run.state == CeleryTaskRunLogState.SUCCEEDED:
                task.update(
                    state=CeleryTaskInvocationLogState.SUCCEEDED,
                    stop=task_last_run.stop,
                    update_only_changed_fields=True
                )
            elif task_last_run and task_last_run.state == CeleryTaskRunLogState.FAILED:
                task.update(
                    state=CeleryTaskInvocationLogState.FAILED,
                    stop=task_last_run.stop,
                    update_only_changed_fields=True
                )
            else:
                try:
                    current_app.tasks[task.name].expire_invocation(
                        task.id,
                        json.loads(task.task_args),
                        json.loads(task.task_kwargs),
                        dict(
                            start=task.start,
                            celery_task_id=task.celery_task_id,
                            stop=now(),
                            name=task.name,
                            queue_name=task.queue_name,
                        )
                    )
                except NotRegistered:
                    task.update(
                        state=CeleryTaskInvocationLogState.EXPIRED,
                        stop=now(),
                        update_only_changed_fields=True
                    )

    def clean_logs(self, type, timestamp, backup_path, stdout):
        storage = import_string(settings.BACKUP_STORAGE_CLASS)()

        qs = get_log_model_from_logger_name(type).search().filter(Q('range', stop={'lt': timestamp})).sort('stop')
        step_timestamp = None
        if qs.count() != 0:
            step_timestamp = list(qs[0:1])[0].stop

        while step_timestamp and step_timestamp < timestamp:
            min_timestamp = datetime.combine(step_timestamp, time.min).replace(tzinfo=utc)
            max_timestamp = datetime.combine(step_timestamp, time.max).replace(tzinfo=utc)

            qs_filtered_by_day = qs.filter(Q('range', stop={'gte': min_timestamp, 'lte': max_timestamp}))

            if qs_filtered_by_day.count() != 0:
                stdout.write(
                    2 * ' ' + 'Cleaning logs for date {} ({})'.format(
                        step_timestamp.date(), qs_filtered_by_day.count()
                    )
                )

                if backup_path:
                    for qs_batch in get_querysets_by_batch(qs_filtered_by_day, settings.PURGE_LOG_BACKUP_BATCH):
                        log_file_name = os.path.join(backup_path, str(step_timestamp.date()))
                        if storage.exists('{}.json.gz'.format(log_file_name)):
                            i = 1
                            while storage.exists('{}({}).json.gz'.format(log_file_name, i)):
                                i += 1
                            log_file_name = '{}({})'.format(log_file_name, i)
                        stdout.write(4 * ' ' + 'generating backup file: {}.json.gz'.format(log_file_name))
                        with storage.open('{}.json.gz'.format(log_file_name), 'wb') as f:
                            with TextIOWrapper(gzip.GzipFile(filename='{}.json'.format(log_file_name),
                                                             fileobj=f, mode='wb')) as gzf:
                                json.dump(
                                    lazy_serialize_qs_without_pk(qs_batch), gzf, cls=DjangoJSONEncoder, indent=5
                                )
                stdout.write(4 * ' ' + 'deleting logs')
                qs_filtered_by_day.delete()

            step_timestamp = min_timestamp + timedelta(days=1)



from importlib import import_module

from django.apps import AppConfig


def import_class(path):
    module_name, class_name = path.rsplit('.', 1)
    return getattr(import_module(module_name), class_name)


class SecurityBackend(AppConfig):

    backend_name = None
    writer = None
    reader = None
    registered_writers = {}
    registered_readers = {}

    def ready(self):
        if self.reader:
            self.reader_inst = import_class(self.reader)()
            SecurityBackend.registered_readers[self.backend_name] = self.reader_inst
        if self.writer:
            self.writer_inst = import_class(self.writer)(self.backend_name)
            SecurityBackend.registered_writers[self.backend_name] = self.writer_inst



from io import StringIO

from contextlib import contextmanager

from django.contrib.auth.models import User

from germanium.decorators import data_provider
from germanium.tools import assert_equal, assert_false, assert_is_not_none

from security.management import call_command


TRUNCATION_CHAR = '…'


class BaseTestCaseMixin:

    databases = ['default', 'security']

    @data_provider
    def create_user(self, username='test', email='test@localhost'):
        return User.objects._create_user(username, email, 'test', is_staff=True, is_superuser=True)



def test_call_command(*args, **kwargs):
    call_command(*args, **kwargs, stdout=StringIO(), stderr=StringIO())



from django.contrib.auth.models import User
from django.core.management.base import BaseCommand

from reversion.revisions import create_revision


class Command(BaseCommand):

    def handle(self, **options):
        for i in range(100):
            with create_revision():
                User.objects._create_user(
                    f'test{i}', f'test{i}@localhost', f'test{i}', is_staff=True, is_superuser=True
                )



import responses

from django import get_version
from django.contrib.contenttypes.models import ContentType
from django.contrib.auth.models import User
from django.test import override_settings

from germanium.decorators import data_consumer
from germanium.test_cases.client import ClientTestCase
from germanium.tools import (
    assert_equal, assert_false, assert_http_not_found, assert_http_ok, assert_http_redirect,
    assert_http_too_many_requests, assert_in, assert_is_none, assert_is_not_none, assert_not_in,
    assert_raises, assert_true, assert_equal_model_fields, assert_length_equal, all_eq_obj, not_none_eq_obj
)

from security.config import settings
from security.decorators import log_with_data
from security.enums import CommandState, LoggerName
from security.backends.sql.models import CommandLog as SQLCommandLog
from security.backends.elasticsearch.models import CommandLog as ElasticsearchCommandLog
from security.backends.signals import (
    command_started, command_error, command_finished, command_output_updated, output_request_finished
)
from security.backends.reader import get_logs_related_with_object
from security.backends.testing import capture_security_logs

from .base import BaseTestCaseMixin, TRUNCATION_CHAR, test_call_command


@override_settings(SECURITY_BACKEND_WRITERS={})
class CommandLogTestCase(BaseTestCaseMixin, ClientTestCase):

    def test_command_should_be_logged(self):
        expected_command_started_data = {
            'name': 'test_command',
            'input': 'verbosity=0',
            'is_executed_from_command_line': False,
            'start': all_eq_obj,
        }
        expected_command_finished_data = {
            **expected_command_started_data,
            'stop': all_eq_obj,
        }
        with capture_security_logs() as logged_data:
            test_call_command('test_command', verbosity=0)
            assert_length_equal(logged_data.command_started, 1)
            assert_length_equal(logged_data.command_finished, 1)
            assert_length_equal(logged_data.command_output_updated, 1)
            assert_length_equal(logged_data.command_error, 0)
            assert_equal(logged_data.command_started[0].data, expected_command_started_data)
            assert_equal(logged_data.command_finished[0].data, expected_command_finished_data)

    @override_settings(SECURITY_LOG_STRING_IO_FLUSH_TIMEOUT=0)
    def test_command_log_string_io_flush_timeout_should_changed(self):
        with capture_security_logs() as logged_data:
            test_call_command('test_command')
            assert_length_equal(logged_data.command_output_updated, 21)

    def test_error_command_should_be_logged(self):
        expected_command_started_data = {
            'name': 'test_error_command',
            'input': '',
            'is_executed_from_command_line': False,
            'start': all_eq_obj,
        }
        expected_command_error_data = {
            **expected_command_started_data,
            'error_message': not_none_eq_obj,
            'stop': all_eq_obj,
        }
        with capture_security_logs() as logged_data:
            with assert_raises(RuntimeError):
                test_call_command('test_error_command')
            assert_length_equal(logged_data.command_started, 1)
            assert_length_equal(logged_data.command_finished, 0)
            assert_length_equal(logged_data.command_error, 1)
            assert_equal(logged_data.command_started[0].data, expected_command_started_data)
            assert_equal(logged_data.command_error[0].data, expected_command_error_data)

    @responses.activate
    @data_consumer('create_user')
    def test_command_with_response_should_be_logged_with_parent_data(self, user):
        responses.add(responses.POST, 'http://localhost/test', body='test')
        with log_with_data(related_objects=[user], slug='TEST', extra_data={'test': 'test'}):
            with capture_security_logs() as logged_data:
                test_call_command('test_command_with_response')
                command_logger = logged_data.command[0]
                output_request_logger = logged_data.output_request[0]
                assert_equal(output_request_logger.parent_with_id, command_logger)
                assert_equal(command_logger.slug, 'TEST')
                assert_equal(command_logger.related_objects, {user})
                assert_equal(command_logger.extra_data, {'test': 'test'})
                assert_equal(output_request_logger.slug, 'TEST')
                assert_equal(output_request_logger.related_objects, {user})
                assert_equal(output_request_logger.extra_data, {'test': 'test'})

    @override_settings(SECURITY_BACKEND_WRITERS={'sql'}, SECURITY_BACKEND_READER='sql')
    @data_consumer('create_user')
    def test_command_should_be_logged_in_sql_backend(self, user):
        with log_with_data(related_objects=[user]):
            test_call_command('test_command', verbosity=0)
            assert_equal(SQLCommandLog.objects.count(), 1)
            sql_command_log = SQLCommandLog.objects.get()
            assert_equal_model_fields(
                sql_command_log,
                name='test_command',
                input='verbosity=0',
                is_executed_from_command_line=False,
                time=(sql_command_log.stop - sql_command_log.start).total_seconds(),
                state=CommandState.SUCCEEDED,
                error_message=None,
            )
            assert_is_not_none(sql_command_log.output)
            assert_equal([rel_obj.object for rel_obj in sql_command_log.related_objects.all()], [user])
            assert_equal(get_logs_related_with_object(LoggerName.COMMAND, user), [sql_command_log])

    @override_settings(SECURITY_BACKEND_WRITERS={'elasticsearch'}, SECURITY_BACKEND_READER='elasticsearch',
                       SECURITY_ELASTICSEARCH_AUTO_REFRESH=True)
    @data_consumer('create_user')
    def test_command_should_be_logged_in_elasticsearch_backend(self, user):
        with capture_security_logs() as logged_data:
            with log_with_data(related_objects=[user]):
                test_call_command('test_command', verbosity=0)
                elasticsearch_command_log = ElasticsearchCommandLog.get(
                    id=logged_data.command[0].id
                )
                assert_equal_model_fields(
                    elasticsearch_command_log,
                    name='test_command',
                    input='verbosity=0',
                    is_executed_from_command_line=False,
                    time=(elasticsearch_command_log.stop - elasticsearch_command_log.start).total_seconds(),
                    state=CommandState.SUCCEEDED,
                    error_message=None,
                )
                assert_is_not_none(elasticsearch_command_log.output)
                assert_equal(
                    [rel_obj for rel_obj in elasticsearch_command_log.related_objects],
                    ['default|3|{}'.format(user.id)]
                )
                assert_equal(get_logs_related_with_object(LoggerName.COMMAND, user), [elasticsearch_command_log])

    @responses.activate
    @override_settings(SECURITY_BACKEND_WRITERS={'logging'})
    @data_consumer('create_user')
    def test_command_should_be_logged_in_logging_backend(self, user):
        with capture_security_logs() as logged_data:
            with self.assertLogs('security.command', level='INFO') as cm:
                test_call_command('test_command', verbosity=0)
                command_log = logged_data.command[0]
                assert_equal(
                    cm.output,
                    [
                        f'INFO:security.command:'
                        f'Command "{command_log.id}" '
                        f'with name "test_command" was started',
                        f'INFO:security.command:'
                        f'Command "{command_log.id}" '
                        f'with name "test_command" was successful',
                    ]
                )

    @override_settings(SECURITY_BACKEND_WRITERS={'sql'})
    @data_consumer('create_user')
    def test_error_command_should_be_logged_in_sql_backend(self, user):
        with log_with_data(related_objects=[user]):
            with assert_raises(RuntimeError):
                test_call_command('test_error_command')
            assert_equal(SQLCommandLog.objects.count(), 1)
            sql_command_log = SQLCommandLog.objects.get()
            assert_equal_model_fields(
                sql_command_log,
                name='test_error_command',
                input='',
                is_executed_from_command_line=False,
                time=(sql_command_log.stop - sql_command_log.start).total_seconds(),
                state=CommandState.FAILED,
                output='',
            )
            assert_is_not_none(sql_command_log.error_message)
            assert_equal([rel_obj.object for rel_obj in sql_command_log.related_objects.all()], [user])

    @override_settings(SECURITY_BACKEND_WRITERS={'elasticsearch'})
    @data_consumer('create_user')
    def test_error_command_should_be_logged_in_elasticsearch_backend(self, user):
        with capture_security_logs() as logged_data:
            with log_with_data(related_objects=[user]):
                with assert_raises(RuntimeError):
                    test_call_command('test_error_command')
                elasticsearch_command_log = ElasticsearchCommandLog.get(
                    id=logged_data.command[0].id
                )
                assert_equal_model_fields(
                    elasticsearch_command_log,
                    name='test_error_command',
                    input='',
                    is_executed_from_command_line=False,
                    time=(elasticsearch_command_log.stop - elasticsearch_command_log.start).total_seconds(),
                    state=CommandState.FAILED,
                    output='',
                )
                assert_is_not_none(elasticsearch_command_log.error_message)
                assert_equal(
                    [rel_obj for rel_obj in elasticsearch_command_log.related_objects],
                    ['default|3|{}'.format(user.id)]
                )

    @responses.activate
    @override_settings(SECURITY_BACKEND_WRITERS={'logging'})
    @data_consumer('create_user')
    def test_error_command_should_be_logged_in_logging_backend(self, user):
        with capture_security_logs() as logged_data:
            with self.assertLogs('security.command', level='INFO') as cm:
                with assert_raises(RuntimeError):
                    test_call_command('test_error_command')
                command_log = logged_data.command[0]
                assert_equal(
                    cm.output,
                    [
                        f'INFO:security.command:'
                        f'Command "{command_log.id}" '
                        f'with name "test_error_command" was started',
                        f'ERROR:security.command:'
                        f'Command "{command_log.id}" '
                        f'with name "test_error_command" failed',
                    ]
                )

    def test_data_change_should_be_connected_with_command_log(self):
        with capture_security_logs() as logged_data:
            test_call_command('create_user')
            assert_equal(
                logged_data.command[0].extra_data,
                {'reversion': {'revisions': [{'id': not_none_eq_obj}], 'total_count': 1}}
            )

    def test_only_20_data_changes_should_be_connected_with_command_log(self):
        with capture_security_logs() as logged_data:
            test_call_command('create_users')
            assert_equal(
                logged_data.command[0].extra_data,
                {
                    'reversion': {
                        'revisions': [{'id': not_none_eq_obj} for _ in range(20)],
                        'total_count': 100
                    }
                }
            )

    @override_settings(SECURITY_COMMAND_LOG_EXCLUDED_COMMANDS=('test_command',))
    def test_excluded_command_should_not_be_logged(self):
        with capture_security_logs() as logged_data:
            test_call_command('test_command')
            assert_length_equal(logged_data.command_started, 0)
            assert_length_equal(logged_data.command_finished, 0)
            assert_length_equal(logged_data.command_output_updated, 0)
            assert_length_equal(logged_data.command_error, 0)



from django.core.management.base import BaseCommand


class Command(BaseCommand):

    def handle(self, **options):
        for i in range(10):
            self.stdout.write(f'STDOUT row {i}')
        for i in range(10):
            self.stderr.write(f'STDERR row {i}')



from django.dispatch import Signal


input_request_started = Signal()
input_request_finished = Signal()
input_request_error = Signal()

output_request_started = Signal()
output_request_finished = Signal()
output_request_error = Signal()

command_started = Signal()
command_output_updated = Signal()
command_finished = Signal()
command_error = Signal()

celery_task_invocation_started = Signal()
celery_task_invocation_triggered = Signal()
celery_task_invocation_ignored = Signal()
celery_task_invocation_timeout = Signal()
celery_task_invocation_expired = Signal()

celery_task_run_started = Signal()
celery_task_run_succeeded = Signal()
celery_task_run_failed = Signal()
celery_task_run_retried = Signal()
celery_task_run_output_updated = Signal()



from attrdict import AttrDict

from django.test.utils import override_settings

from security.backends.writer import BaseBackendWriter

from .app import SecurityTestingBackend


class CapturedLog:

    def __init__(self, logger):
        self.id = logger.id
        self.slug = logger.slug
        self.data = dict(logger.data)
        self.related_objects = list(logger.related_objects)
        self.extra_data = dict(logger.extra_data)
        self.logger = logger


class capture_security_logs(override_settings):

    logged_data = None
    _receivers = None

    def __init__(self, set_testing_writer=False):
        kwargs = {}
        if set_testing_writer:
            kwargs['SECURITY_BACKEND_READER'] = SecurityTestingBackend.backend_name
        super().__init__(**kwargs)

    def _get_receiver(self, signal_name, use_wrapper=False):
        def _log_receiver(sender, logger, signal, **kwargs):
            if use_wrapper:
                logger = CapturedLog(logger)
            capture_security_logs.logged_data[signal_name].append(logger)
        return _log_receiver

    def _set_signal_receiver(self, signal_name, signal, use_wrapper=False):
        receiver = self._get_receiver(signal_name, use_wrapper)
        capture_security_logs._receivers[signal_name] = (receiver, signal)
        capture_security_logs.logged_data[signal_name] = []
        signal.connect(receiver, weak=True)

    def enable(self):
        super().enable()
        capture_security_logs.logged_data = AttrDict()
        capture_security_logs._receivers = {}

        for signal_name, signal in BaseBackendWriter.CAPTURED_SIGNALS.items():
            if signal_name.endswith('_started'):
                self._set_signal_receiver(signal_name[0:-8], signal)
            self._set_signal_receiver(signal_name, signal, use_wrapper=True)
        return capture_security_logs.logged_data

    def disable(self):
        for receiver, signal in capture_security_logs._receivers.values():
            signal.disconnect(receiver)
        super().disable()



from settings.base import *  # pylint: disable=E0401


DEBUG = TEMPLATE_DEBUG = THUMBNAIL_DEBUG = True

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(PROJECT_DIR, 'var', 'db', 'sqlite.db'),
        'USER': '',
        'PASSWORD': '',
    },
    'security': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(PROJECT_DIR, 'var', 'db', 'log.db'),
        'USER': '',
        'PASSWORD': '',
    },
}

ROOT_URLCONF = 'urls'

STATIC_ROOT = ''

# Additional locations of static files
STATICFILES_DIRS = (
    STATICFILES_ROOT,
)

SECURITY_LOG_DB_NAME = 'security'
DATABASE_ROUTERS = ['security.backends.sql.db_router.MultipleDBSecurityLoggerRouter']

ALLOWED_HOSTS = ['*']

SECURITY_BACKENDS = ['sql']



import os, shutil

import responses
from datetime import timedelta

from freezegun import freeze_time

from django.conf import settings as django_settings
from django.test import override_settings
from django.utils.timezone import now

from germanium.test_cases.client import ClientTestCase
from germanium.tools import (
    assert_equal, assert_false, assert_http_not_found, assert_http_ok, assert_http_redirect,
    assert_http_too_many_requests, assert_in, assert_is_none, assert_is_not_none, assert_not_in,
    assert_raises, assert_true, assert_equal_model_fields
)

from security import requests
from security.config import settings
from security.backends.sql.models import CommandLog as SQLCommandLog
from security.backends.elasticsearch.models import CommandLog as ElasticsearchCommandLog
from security.backends.sql.models import CeleryTaskRunLog as SQLCeleryTaskRunLog
from security.backends.sql.models import CeleryTaskInvocationLog as SQLCeleryTaskInvocationLog
from security.backends.elasticsearch.models import CeleryTaskRunLog as ElasticsearchCeleryTaskRunLog
from security.backends.elasticsearch.models import CeleryTaskInvocationLog as ElasticsearchCeleryTaskInvocationLog
from security.backends.sql.models import InputRequestLog as SQLInputRequestLog
from security.backends.elasticsearch.models import InputRequestLog as ElasticsearchInputRequestLog
from security.backends.sql.models import OutputRequestLog as SQLOutputRequestLog
from security.backends.elasticsearch.models import OutputRequestLog as ElasticsearchOutputRequestLog

from security.backends.testing import capture_security_logs

from apps.test_security.tasks import sum_task

from .base import BaseTestCaseMixin, test_call_command


@override_settings(SECURITY_BACKEND_WRITERS={})
class CommandTestCase(BaseTestCaseMixin, ClientTestCase):

    SQL_LOG_MODELS = {
        'input-request': SQLInputRequestLog,
        'output-request': SQLOutputRequestLog,
        'command': SQLCommandLog,
        'celery-invocation': SQLCeleryTaskInvocationLog,
        'celery-run': SQLCeleryTaskRunLog,
    }

    @responses.activate
    @override_settings(
        SECURITY_BACKEND_WRITERS={'sql'},
        SECURITY_COMMAND_LOG_EXCLUDED_COMMANDS={'purge_logs'},
        SECURITY_BACKUP_STORAGE_PATH=os.path.join(django_settings.PROJECT_DIR, 'var', 'backup_sql')
    )
    def test_sql_purge_logs_should_remove_logged_data(self):
        responses.add(responses.GET, 'https://localhost/test', body='test')

        test_call_command('test_command')
        sum_task.apply(args=(5, 8))
        assert_http_ok(self.get('/home/'))
        requests.get('https://localhost/test')

        sql_type_model = {
            ('input-request', SQLInputRequestLog),
            ('output-request', SQLOutputRequestLog),
            ('command', SQLCommandLog),
            ('celery-task-invocation', SQLCeleryTaskInvocationLog),
            ('celery-task-run', SQLCeleryTaskRunLog),
        }

        for log_type, log_model in sql_type_model:
            log_directory = os.path.join(settings.BACKUP_STORAGE_PATH, log_type)
            os.makedirs(log_directory)
            test_call_command('purge_logs', type=log_type, interactive=False, expiration='1d', backup=log_type)
            assert_equal(log_model.objects.count(), 1)

            with freeze_time(now() + timedelta(days=1, minutes=1)):
                test_call_command('purge_logs', type=log_type, interactive=False, expiration='1d', backup=log_type)
                assert_equal(log_model.objects.count(), 0)
            assert_equal(len(os.listdir(log_directory)), 1)
        shutil.rmtree(settings.BACKUP_STORAGE_PATH)

    @responses.activate
    @override_settings(
        SECURITY_BACKEND_WRITERS={'elasticsearch'},
        SECURITY_COMMAND_LOG_EXCLUDED_COMMANDS={'purge_logs'},
        SECURITY_BACKUP_STORAGE_PATH=os.path.join(django_settings.PROJECT_DIR, 'var', 'backup_elastic')
    )
    def test_elasticsearch_purge_logs_should_remove_logged_data(self):
        responses.add(responses.GET, 'https://localhost/test', body='test')
        with capture_security_logs() as logged_data:
            test_call_command('test_command')
            sum_task.apply(args=(5, 8))
            assert_http_ok(self.get('/home/'))
            requests.get('https://localhost/test')

            elasticsearch_type_model_receivers = (
                ('input-request', ElasticsearchInputRequestLog, 'input_request'),
                ('output-request', ElasticsearchOutputRequestLog, 'output_request'),
                ('command', ElasticsearchCommandLog, 'command'),
                ('celery-task-invocation', ElasticsearchCeleryTaskInvocationLog, 'celery_task_invocation'),
                ('celery-task-run', ElasticsearchCeleryTaskRunLog, 'celery_task_run'),
            )

            for log_type, log_model, log_data_name in elasticsearch_type_model_receivers:
                log_directory = os.path.join(settings.BACKUP_STORAGE_PATH, log_type)
                os.makedirs(log_directory)
                test_call_command('purge_logs', type=log_type, interactive=False, expiration='1d', backup=log_type)
                log_model._index.refresh()
                assert_equal(log_model.search().filter(
                    'ids', values=[logged_data[log_data_name][0].id]
                ).count(), 1)

                with freeze_time(now() + timedelta(days=1, minutes=1)):
                    test_call_command('purge_logs', type=log_type, interactive=False, expiration='1d', backup=log_type)
                    log_model._index.refresh()
                    assert_equal(log_model.search().filter(
                        'ids', values=[logged_data[log_data_name][0].id]
                    ).count(), 0)
                assert_equal(len(os.listdir(log_directory)), 1)
        shutil.rmtree(settings.BACKUP_STORAGE_PATH)



from elasticsearch import Elasticsearch, RequestsHttpConnection
from requests_aws4auth import AWS4Auth
import boto3

host = 'vpc-mallpay-be-accept-fyaqep5vxdjqrkmrqbbem2nimq.eu-central-1.es.amazonaws.com'
region = 'eu-central-1'

service = 'es'
print('init boto 3')

credentials = boto3.Session().get_credentials()
print('boto')
awsauth = AWS4Auth(credentials.access_key, credentials.secret_key, region, service, session_token=credentials.token)
search = Elasticsearch(
    hosts = [{'host': host, 'port': 443}],
    http_auth = awsauth,
    use_ssl = True,
    verify_certs = True,
    connection_class = RequestsHttpConnection
)
print(search)

document = {
    "title": "Moneyball",
    "director": "Bennett Miller",
    "year": "2011"
}

search.index(index="movies", doc_type="_doc", id="5", body=document)

print(search.get(index="movies", doc_type="_doc", id="5"))



from django.core.management.base import BaseCommand, CommandError

from security.backends.elasticsearch.connection import set_connection
from security.backends.elasticsearch.models import (
    InputRequestLog, OutputRequestLog, CommandLog, CeleryTaskRunLog, CeleryTaskInvocationLog
)


class Command(BaseCommand):

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument(
            '--noinput', '--no-input', action='store_false', dest='interactive',
            help='Tells Django to NOT prompt the user for input of any kind.',
        )

    def handle(self, **options):
        if options['interactive']:
            message = (
                'This will delete existing logs!\n'
                'Are you sure you want to do this?\n\n'
                "Type 'yes' to continue, or 'no' to cancel: "
            )
            if input(message) != 'yes':
                raise CommandError('Init elasticsearch logs cancelled.')

        self.stdout.write('Init elasticsearch logs')
        set_connection(init_documents=False)
        for document in InputRequestLog, OutputRequestLog, CommandLog, CeleryTaskRunLog, CeleryTaskInvocationLog:
            document._index.delete(ignore=404)
            document.init()



import sys

import traceback

from security.config import settings
from security.enums import LoggerName
from security.logging.common import get_last_logger


def get_command_stream():
    last_command_logger = get_last_logger(LoggerName.COMMAND)
    return last_command_logger.stream if last_command_logger is not None else sys.stdout


class CommandExecutor:
    """
    A helper class that runs a django command and logs details about its run into DB.
    """

    def __init__(self, command_function, command_args=None, command_kwargs=None, stdout=None, stderr=None, **kwargs):
        """
        Initializes the command logger.

        Arguments:
            command_function: Callable that implements the command.
            stdout: Custom stream where command's standard output will be written.
            stderr: Custom stream where command's error output will be written.
            **kwargs: Keyword arguments passed to CommandLog model.
        """
        assert 'name' in kwargs, 'Key name must be provided in kwargs'

        self.command_function = command_function
        self.command_args = command_args or ()
        self.command_kwargs = command_kwargs or {}

        self.kwargs = kwargs
        self.stdout = stdout if stdout else sys.stdout
        self.stderr = stderr if stderr else sys.stderr

    def run(self):
        """
        Runs the command function and returns its return value or re-raises any exceptions. The run of the command will
        not be logged if it is in excluded commands setting.
        """
        from security.logging.commands.logger import CommandLogger
        from security.utils import LogStringIO, TeeStringIO

        if self.kwargs['name'] in settings.COMMAND_LOG_EXCLUDED_COMMANDS:
            return self.command_function(
                stdout=self.stdout, stderr=self.stderr, *self.command_args, **self.command_kwargs
            )

        with CommandLogger() as command_logger:
            command_logger.log_started(**self.kwargs)
            self.output = LogStringIO(
                flush_callback=lambda output_stream: command_logger.log_output_updated(output_stream.getvalue())
            )
            command_logger.stream = self.output
            stdout = TeeStringIO(self.stdout, self.output)
            stderr = TeeStringIO(self.stderr, self.output)
            try:
                ret_val = self.command_function(
                    stdout=stdout, stderr=stderr, *self.command_args, **self.command_kwargs
                )
                command_logger.log_finished()
                return ret_val
            except (Exception, KeyboardInterrupt, SystemExit) as ex:  # pylint: disable=W0703
                command_logger.log_exception(traceback.format_exc())
                raise ex
            finally:
                self.output.close()



import json

import responses

from django import get_version
from django.test import override_settings
from django.utils.encoding import force_text

from germanium.decorators import data_consumer
from germanium.test_cases.client import ClientTestCase
from germanium.tools import (
    assert_equal, assert_false, assert_http_not_found, assert_http_ok, assert_http_redirect,
    assert_http_too_many_requests, assert_in, assert_is_none, assert_is_not_none, assert_not_in,
    assert_raises, assert_true, assert_equal_model_fields, assert_length_equal, all_eq_obj
)

from security.enums import RequestLogState
from security.config import settings
from security.decorators import log_with_data
from security.backends.sql.models import InputRequestLog as SQLInputRequestLog
from security.backends.elasticsearch.models import InputRequestLog as ElasticsearchInputRequestLog

from security import requests
from security.backends.signals import (
    input_request_started, input_request_finished, input_request_error, output_request_started, output_request_finished
)
from security.backends.testing import capture_security_logs

from .base import BaseTestCaseMixin, TRUNCATION_CHAR


@override_settings(SECURITY_BACKEND_WRITERS={}, SECURITY_LOG_RESPONSE_BODY_CONTENT_TYPES=None)
class InputRequestLogTestCase(BaseTestCaseMixin, ClientTestCase):

    def test_input_request_to_homepage_should_be_logged(self):
        expected_input_request_started_data = {
            'request_headers': {'COOKIE': '[Filtered]'},
            'request_body': '',
            'user_id': None,
            'method': 'GET',
            'host': 'testserver',
            'path': '/home/',
            'queries': {'name': 'value'},
            'is_secure': False,
            'ip': '127.0.0.1',
            'start': all_eq_obj,
            'view_slug': 'home'
        }
        expected_input_request_finished_data = {
            **expected_input_request_started_data,
            'stop': all_eq_obj,
            'response_code': 200,
            'response_headers': {'Content-Type': 'text/html; charset=utf-8', 'X-Frame-Options': 'DENY'},
            'response_body': 'home page response',
        }
        with capture_security_logs() as logged_data:
            assert_http_ok(self.get('/home/?name=value'))
            assert_length_equal(logged_data.input_request_started, 1)
            assert_length_equal(logged_data.input_request_finished, 1)
            assert_length_equal(logged_data.input_request_error, 0)
            assert_equal(logged_data.input_request_started[0].data, expected_input_request_started_data)
            assert_equal(logged_data.input_request_finished[0].data, expected_input_request_finished_data)

    @data_consumer('create_user')
    def test_input_request_to_login_page_should_be_logged(self, user):
        expected_input_request_started_data = {
            'request_headers': {'COOKIE': '[Filtered]'},
            'request_body': (
                '--BoUnDaRyStRiNg\r\n'
                'Content-Disposition: form-data; name="username"\r\n'
                '\r\n'
                'test\r\n'
                '--BoUnDaRyStRiNg\r\n'
                '[Filtered]\n'
                '--BoUnDaRyStRiNg--\r\n'
            ),
            'user_id': None,
            'method': 'POST',
            'host': 'testserver',
            'path': '/admin/login/',
            'queries': {},
            'is_secure': False,
            'ip': '127.0.0.1',
            'start': all_eq_obj,
            'view_slug': 'admin:login'
        }
        expected_input_request_finished_data = {
            **expected_input_request_started_data,
            'stop': all_eq_obj,
            'response_code': 302,
            'response_headers': {
                'Cache-Control': 'max-age=0, no-cache, no-store, must-revalidate, private',
                'Content-Type': 'text/html; charset=utf-8',
                'Expires': all_eq_obj,
                'Location': '/accounts/profile/',
                'Vary': 'Cookie',
                'X-Frame-Options': 'DENY'
            },
            'response_body': '',
            'user_id': user.pk,
        }

        with capture_security_logs() as logged_data:
            assert_http_redirect(self.post('/admin/login/', data={'username': 'test', 'password': 'test'}))
            assert_length_equal(logged_data.input_request_started, 1)
            assert_length_equal(logged_data.input_request_finished, 1)
            assert_equal(logged_data.input_request_started[0].data, expected_input_request_started_data)
            assert_equal(logged_data.input_request_finished[0].data, expected_input_request_finished_data)

    def test_input_request_to_error_page_should_be_logged(self):
        expected_input_request_started_data = {
            'request_headers': {'COOKIE': '[Filtered]'},
            'request_body': '',
            'user_id': None,
            'method': 'GET',
            'host': 'testserver',
            'path': '/error/',
            'queries': {},
            'is_secure': False,
            'ip': '127.0.0.1',
            'start': all_eq_obj,
            'view_slug': 'apps.test_security.views.error_view'
        }
        expected_input_request_error_data = {
            **expected_input_request_started_data,
            'error_message': all_eq_obj,
        }
        expected_input_request_finished_data = {
            **expected_input_request_error_data,
            'stop': all_eq_obj,
            'response_code': 500,
            'response_headers': all_eq_obj,
            'response_body': all_eq_obj,
        }

        with capture_security_logs() as logged_data:
            with assert_raises(RuntimeError):
                assert_http_ok(self.get('/error/'))
            assert_length_equal(logged_data.input_request_started, 1)
            assert_length_equal(logged_data.input_request_finished, 1)
            assert_length_equal(logged_data.input_request_error, 1)
            assert_equal(logged_data.input_request_started[0].data, expected_input_request_started_data)
            assert_equal(logged_data.input_request_finished[0].data, expected_input_request_finished_data)
            assert_equal(logged_data.input_request_error[0].data, expected_input_request_error_data)

    @override_settings(SECURITY_LOG_REQUEST_IGNORE_IP=('127.0.0.1',))
    def test_ignored_client_ip_should_not_be_logged(self):
        with capture_security_logs() as logged_data:
            assert_http_ok(self.get('/home/'))
            assert_length_equal(logged_data.input_request, 0)

    @override_settings(SECURITY_LOG_REQUEST_IGNORE_URL_PATHS=('/home/',))
    def test_ignored_client_ip_should_not_be_logged(self):
        with capture_security_logs() as logged_data:
            assert_http_ok(self.get('/home/'))
            assert_length_equal(logged_data.input_request, 0)

    @override_settings(SECURITY_LOG_REQUEST_BODY_LENGTH=10)
    def test_request_body_should_be_truncated(self):
        with capture_security_logs() as logged_data:
            self.post('/admin/login/', data={'username': 20 * 'a', 'password': 20 * 'b'})
            assert_equal(len(logged_data.input_request[0].data['request_body']), 10)
            assert_true(logged_data.input_request[0].data['request_body'].endswith(TRUNCATION_CHAR))

    @override_settings(SECURITY_LOG_RESPONSE_BODY_LENGTH=10)
    def test_response_body_should_be_truncated(self):
        with capture_security_logs() as logged_data:
            self.post('/admin/login/', data={'username': 20 * 'a', 'password': 20 * 'b'})
            assert_equal(len(logged_data.input_request[0].data['response_body']), 10)
            assert_true(logged_data.input_request[0].data['response_body'].endswith(TRUNCATION_CHAR))

    @override_settings(SECURITY_LOG_REQUEST_BODY_LENGTH=None, SECURITY_HIDE_SENSITIVE_DATA=False)
    def test_request_body_truncation_should_be_turned_off(self):
        with capture_security_logs() as logged_data:
            self.post('/admin/login/', data={'username': 2000 * 'a', 'password': 2000 * 'b'})
            assert_equal(len(logged_data.input_request[0].data['request_body']), 4162)
            assert_false(logged_data.input_request[0].data['request_body'].endswith(TRUNCATION_CHAR))

    @override_settings(SECURITY_LOG_RESPONSE_BODY_LENGTH=None, SECURITY_HIDE_SENSITIVE_DATA=False)
    def test_response_body_truncation_should_be_turned_off(self):
        with capture_security_logs() as logged_data:
            response = self.post('/admin/login/', data={'username': 2000 * 'a', 'password': 2000 * 'b'})
            assert_equal(logged_data.input_request[0].data['response_body'], force_text(response.content))
            assert_false(logged_data.input_request[0].data['response_body'].endswith(TRUNCATION_CHAR))

    @override_settings(SECURITY_LOG_RESPONSE_BODY_CONTENT_TYPES=())
    def test_not_allowed_content_type_body_should_not_be_logged(self):
        with capture_security_logs() as logged_data:
            assert_http_ok(self.get('/home/'))
            assert_is_none(logged_data.input_request[0].data['response_body'])

    @override_settings(SECURITY_LOG_REQUEST_BODY_LENGTH=100, SECURITY_LOG_JSON_STRING_LENGTH=10)
    def test_json_request_data_should_be_truncated(self):
        with capture_security_logs() as logged_data:
            self.c.post('/admin/login/', data=json.dumps({'a': 50 * 'a', 'b': 50 * 'b'}),
                        content_type='application/json')
            assert_equal(
                json.loads(logged_data.input_request[0].data['request_body']),
                json.loads('{"a": "%s%s", "b": "%s%s"}' % (
                    9 * 'a', TRUNCATION_CHAR, 9 * 'b', TRUNCATION_CHAR
                ))
            )

    @override_settings(SECURITY_LOG_REQUEST_BODY_LENGTH=50, SECURITY_LOG_JSON_STRING_LENGTH=None)
    def test_json_request_data_should_not_be_truncated(self):
        with capture_security_logs() as logged_data:
            self.c.post('/admin/login/', data=json.dumps({'a': 50 * 'a'}),
                        content_type='application/json')
            assert_equal(logged_data.input_request[0].data['request_body'], '{"a": "' + 42 * 'a' + TRUNCATION_CHAR)
            assert_true(logged_data.input_request[0].data['request_body'].endswith(TRUNCATION_CHAR))

    @override_settings(SECURITY_LOG_REQUEST_BODY_LENGTH=100, SECURITY_LOG_JSON_STRING_LENGTH=10)
    def test_json_request_should_have_been_whole_truncated_values_and_inner_data_too(self):
        with capture_security_logs() as logged_data:
            self.c.post('/admin/login/', data=json.dumps({50 * 'a': 50 * 'a', 50 * 'b': 50 * 'b'}),
                        content_type='application/json')
            assert_equal(len(logged_data.input_request[0].data['request_body']), 100)
            assert_true(logged_data.input_request[0].data['request_body'].endswith(TRUNCATION_CHAR))

    @responses.activate
    def test_output_request_should_be_logged_with_input_request(self):
        with capture_security_logs() as logged_data:
            responses.add(responses.GET, 'http://localhost', body='test')
            assert_equal(self.get('/proxy/?url=http://localhost').content, b'test')
            assert_length_equal(logged_data.output_request_started, 1)
            assert_length_equal(logged_data.output_request_finished, 1)
            assert_equal(
                logged_data.output_request[0].parent_with_id,
                logged_data.input_request[0]
            )

    @responses.activate
    @data_consumer('create_user')
    def test_output_logged_request_should_be_related_with_object(self, user):
        with capture_security_logs() as logged_data:
            assert_http_redirect(self.post('/admin/login/', data={'username': 'test', 'password': 'test'}))
            responses.add(responses.GET, 'http://localhost', body='test')
            assert_equal(self.get('/proxy/?url=http://localhost').content, b'test')
            assert_equal(len(logged_data.output_request[0].related_objects), 1)
            assert_equal(list(logged_data.output_request[0].related_objects)[0], user)

    @responses.activate
    @data_consumer('create_user')
    def test_input_logged_request_should_have_set_data(self, user):
        with capture_security_logs() as logged_data:
            assert_http_redirect(self.post('/admin/login/', data={'username': 'test', 'password': 'test'}))
            assert_http_ok(self.get('/home/'))
            assert_equal(len(logged_data.input_request[1].related_objects), 1)
            assert_equal(list(logged_data.input_request[1].related_objects)[0], user)
            assert_equal(logged_data.input_request_finished[1].slug, 'user-home')

    @data_consumer('create_user')
    @override_settings(SECURITY_SENSITIVE_DATA_REPLACEMENT='(Filtered)')
    def test_sensitive_replacement_should_be_changed(self, user):
        with capture_security_logs() as logged_data:
            assert_http_redirect(self.post('/admin/login/', data={'username': 'test', 'password': 'test'}))
            assert_equal(logged_data.input_request[0].data['request_headers']['COOKIE'], '(Filtered)')

    def test_sensitive_queries_should_be_hidden(self):
        with capture_security_logs() as logged_data:
            assert_http_ok(self.get('/home/?token=test'))
            assert_equal(logged_data.input_request[0].data['queries']['token'], '[Filtered]')

    @data_consumer('create_user')
    def test_sensitive_headers_should_be_hidden(self, user):
        with capture_security_logs() as logged_data:
            assert_http_redirect(self.post('/admin/login/', data={'username': 'test', 'password': 'test'}))
            assert_equal(logged_data.input_request[0].data['request_headers']['COOKIE'], '[Filtered]')

    def test_sensitive_data_body_in_raw_form_should_be_hidden(self):
        with capture_security_logs() as logged_data:
            self.post('/admin/login/', data={'username': 'test', 'password': 'secret-password\nddd'})
            assert_in('[Filtered]', logged_data.input_request[0].data['request_body'])

    def test_sensitive_data_body_in_json_should_be_hidden(self):
        with capture_security_logs() as logged_data:
            self.c.post('/admin/login/', data=json.dumps({'username': 'test', 'password': 'secret-password'}),
                        content_type='application/json')
            assert_in('"password": "[Filtered]"', logged_data.input_request[0].data['request_body'])
            assert_not_in(
                '"password": "secret-password"', logged_data.input_request[0].data['request_body']
            )

    @override_settings(SECURITY_BACKEND_WRITERS={'sql'})
    @data_consumer('create_user')
    def test_input_request_to_homepage_should_be_logged_in_sql_backend(self, user):
        with log_with_data(related_objects=[user]):
            assert_http_ok(self.get('/home/?name=value'))
            assert_equal(SQLInputRequestLog.objects.count(), 1)
            sql_input_request_log = SQLInputRequestLog.objects.get()
            assert_equal_model_fields(
                sql_input_request_log,
                request_headers={'COOKIE': '[Filtered]'},
                request_body='',
                user_id=None,
                method='GET',
                host='testserver',
                path='/home/',
                queries={'name': 'value'},
                is_secure=False,
                ip='127.0.0.1',
                view_slug='home',
                slug=None,
                time=(sql_input_request_log.stop - sql_input_request_log.start).total_seconds(),
                extra_data={},
                error_message=None,
                response_code=200,
                response_headers={'Content-Type': 'text/html; charset=utf-8', 'X-Frame-Options': 'DENY'},
                response_body='home page response',
                state=RequestLogState.INFO,
            )
            assert_equal([rel_obj.object for rel_obj in sql_input_request_log.related_objects.all()], [user])

    @override_settings(SECURITY_BACKEND_WRITERS={'elasticsearch'})
    @data_consumer('create_user')
    def test_input_request_to_homepage_should_be_logged_in_elasticsearch_backend(self, user):
        with log_with_data(related_objects=[user]):
            with capture_security_logs() as logged_data:
                assert_http_ok(self.get('/home/?name=value'))
                elasticsearch_input_request_log = ElasticsearchInputRequestLog.get(
                    id=logged_data.input_request[0].id
                )
                assert_equal_model_fields(
                    elasticsearch_input_request_log,
                    request_headers='{"COOKIE": "[Filtered]"}',
                    request_body='',
                    user_id=None,
                    method='GET',
                    host='testserver',
                    path='/home/',
                    queries='{"name": "value"}',
                    is_secure=False,
                    ip='127.0.0.1',
                    view_slug='home',
                    slug=None,
                    time=(elasticsearch_input_request_log.stop - elasticsearch_input_request_log.start).total_seconds(),
                    extra_data={},
                    error_message=None,
                    response_code=200,
                    response_headers='{"Content-Type": "text/html; charset=utf-8", "X-Frame-Options": "DENY"}',
                    response_body='home page response',
                    state=RequestLogState.INFO,
                )
                assert_equal(
                    [rel_obj for rel_obj in elasticsearch_input_request_log.related_objects],
                    ['default|3|{}'.format(user.id)]
                )

    @override_settings(SECURITY_BACKEND_WRITERS={'logging'})
    def test_input_request_to_homepage_should_be_logged_in_logging_backend(self):
        with capture_security_logs() as logged_data:
            with self.assertLogs('security.input_request', level='INFO') as cm:
                assert_http_ok(self.get('/home/?name=value'))
                assert_equal(
                    cm.output,
                    [
                        f'INFO:security.input_request:'
                        f'Input request "{logged_data.input_request[0].id}" '
                        f'to "testserver" with path "/home/" was started',
                        f'INFO:security.input_request:'
                        f'Input request "{logged_data.input_request[0].id}" '
                        f'to "testserver" with path "/home/" was finished'
                    ]
                )

    @override_settings(SECURITY_BACKEND_WRITERS={'sql'})
    def test_input_request_to_error_page_should_be_logged_in_sql_backend(self):
        with assert_raises(RuntimeError):
            self.get('/error/')
        assert_equal(SQLInputRequestLog.objects.count(), 1)
        sql_input_request_log = SQLInputRequestLog.objects.get()
        assert_equal_model_fields(
            sql_input_request_log,
            method='GET',
            path='/error/',
            time=(sql_input_request_log.stop - sql_input_request_log.start).total_seconds(),
            response_code=500,
            state=RequestLogState.ERROR,
        )
        assert_is_not_none(sql_input_request_log.error_message)

    @override_settings(SECURITY_BACKEND_WRITERS={'elasticsearch'})
    def test_input_request_to_error_page_should_be_logged_in_elasticsearch_backend(self):
        with capture_security_logs() as logged_data:
            with assert_raises(RuntimeError):
                self.get('/error/')
            elasticsearch_input_request_log = ElasticsearchInputRequestLog.get(
                id=logged_data.input_request[0].id
            )
            assert_equal_model_fields(
                elasticsearch_input_request_log,
                method='GET',
                path='/error/',
                time=(elasticsearch_input_request_log.stop - elasticsearch_input_request_log.start).total_seconds(),
                response_code=500,
                state=RequestLogState.ERROR,
            )
            assert_is_not_none(elasticsearch_input_request_log.error_message)

    @override_settings(SECURITY_BACKEND_WRITERS={'logging'})
    def test_input_request_to_error_page_should_be_logged_in_logging_backend(self):
        with capture_security_logs() as logged_data:
            with self.assertLogs('security.input_request', level='INFO') as cm:
                with assert_raises(RuntimeError):
                    self.get('/error/')
                assert_equal(
                    cm.output,
                    [
                        f'INFO:security.input_request:'
                        f'Input request "{logged_data.input_request[0].id}" '
                        f'to "testserver" with path "/error/" was started',
                        f'ERROR:security.input_request:'
                        f'Input request "{logged_data.input_request[0].id}" '
                        f'to "testserver" with path "/error/" failed',
                        f'INFO:security.input_request:'
                        f'Input request "{logged_data.input_request[0].id}" '
                        f'to "testserver" with path "/error/" was finished'
                    ]
                )

    @override_settings(SECURITY_BACKEND_WRITERS={'sql'})
    def test_input_request_to_404_page_should_be_logged_in_sql_backend(self):
        assert_http_not_found(self.get('/404/'))
        assert_equal(SQLInputRequestLog.objects.count(), 1)
        sql_input_request_log = SQLInputRequestLog.objects.get()
        assert_equal_model_fields(
            sql_input_request_log,
            response_code=404,
            state=RequestLogState.WARNING,
        )

    @override_settings(SECURITY_BACKEND_WRITERS={'elasticsearch'})
    def test_input_request_to_404_page_should_be_logged_in_elasticsearch_backend(self):
        with capture_security_logs() as logged_data:
            assert_http_not_found(self.get('/404/'))
            elasticsearch_input_request_log = ElasticsearchInputRequestLog.get(
                id=logged_data.input_request[0].id
            )
            assert_equal_model_fields(
                elasticsearch_input_request_log,
                response_code=404,
                state=RequestLogState.WARNING,
            )

    def test_decorated_view_with_hide_request_body_should_not_log_request_body(self):
        with capture_security_logs() as logged_data:
            self.post('/hide-request-body/', data={'a': 20 * 'a', 'b': 20 * 'b'})
            assert_equal(logged_data.input_request[0].data['request_body'], '[Filtered]')

    def test_decorated_view_with_log_exempt_should_not_log_request(self):
        with capture_security_logs() as logged_data:
            self.get('/log-exempt/')
            assert_length_equal(logged_data.input_request, 0)

    def test_throttling_should_not_be_raised(self):
        for _ in range(100):
            assert_http_redirect(self.get('/admin/'))

    @override_settings(SECURITY_BACKEND_WRITERS={'sql'}, SECURITY_BACKEND_READER='sql')
    @override_settings(SECURITY_DEFAULT_THROTTLING_VALIDATORS_PATH='apps.test_security.tests.throttling_validators')
    def test_throttling_configuration_with_sql_backend_should_return_429(self):
        for _ in range(2):
            assert_http_redirect(self.get('/admin/'))
        assert_http_too_many_requests(self.get('/admin/'))

    @override_settings(SECURITY_BACKEND_WRITERS={'elasticsearch'}, SECURITY_BACKEND_READER='elasticsearch')
    @override_settings(
        SECURITY_DEFAULT_THROTTLING_VALIDATORS_PATH='apps.test_security.tests.throttling_validators'
    )
    def test_throttling_configuration_with_elasticsearch_backend_should_return_429(self):
        for _ in range(2):
            assert_http_redirect(self.get('/admin/'))
            ElasticsearchInputRequestLog._index.refresh()
        assert_http_too_many_requests(self.get('/admin/'))

    @override_settings(SECURITY_BACKEND_WRITERS={'testing'}, SECURITY_BACKEND_READER='testing')
    @override_settings(
        SECURITY_DEFAULT_THROTTLING_VALIDATORS_PATH='apps.test_security.tests.throttling_validators'
    )
    @capture_security_logs()
    def test_throttling_configuration_with_testing_backends_should_return_429(self):
        for _ in range(2):
            assert_http_redirect(self.get('/admin/'))
            ElasticsearchInputRequestLog._index.refresh()
        assert_http_too_many_requests(self.get('/admin/'))

    @override_settings(SECURITY_BACKEND_WRITERS={'sql'}, SECURITY_BACKEND_READER='sql')
    @override_settings(SECURITY_DEFAULT_THROTTLING_VALIDATORS_PATH='apps.test_security.tests.throttling_validators')
    def test_decorated_view_with_throttling_exempt_should_not_raise_throttling_exception(self):
        for _ in range(20):
            assert_http_ok(self.get('/throttling-exempt/'))
        assert_http_ok(self.get('/throttling-exempt/'))

    @override_settings(SECURITY_BACKEND_WRITERS={'sql'}, SECURITY_BACKEND_READER='sql')
    def test_decorated_view_with_throttling_should_raise_throttling_exception(self):
        assert_http_ok(self.get('/extra-throttling/'))
        assert_http_too_many_requests(self.get('/extra-throttling/'))

    @data_consumer('create_user')
    def test_slug_and_related_data_should_be_send_to_input_request_logger(self, user):
        with log_with_data(related_objects=[user], slug='TEST'):
            with capture_security_logs() as logged_data:
                assert_http_ok(self.get('/home/'))
                assert_equal(logged_data.input_request[0].related_objects, {user})
                assert_equal(logged_data.input_request[0].slug, 'TEST')



import os
import sys

from django.conf import settings

from django_celery_extensions.celery import Celery


app = Celery('example', task_cls='security.task:LoggedTask')

# Using a string here means the worker will not have to
# pickle the object when using Windows.
app.config_from_object('django.conf:settings')
app.autodiscover_tasks()



# Generated by Django 3.1.13 on 2021-09-06 14:47

import django.core.serializers.json
from django.db import migrations, models
import django.db.models.deletion
import enumfields.fields
import security.enums


class Migration(migrations.Migration):
    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='CeleryTaskInvocationLog',
            fields=[
                ('id', models.UUIDField(primary_key=True, serialize=False)),
                ('extra_data',
                 models.JSONField(blank=True, encoder=django.core.serializers.json.DjangoJSONEncoder, null=True)),
                ('slug', models.CharField(blank=True, db_index=True, max_length=255, null=True)),
                ('start', models.DateTimeField(editable=False)),
                ('stop', models.DateTimeField(blank=True, editable=False, null=True)),
                ('time', models.FloatField(blank=True, null=True)),
                ('celery_task_id', models.UUIDField(blank=True, db_index=True, null=True)),
                ('name', models.CharField(db_index=True, max_length=250)),
                ('queue_name', models.CharField(blank=True, max_length=250, null=True)),
                ('applied_at', models.DateTimeField()),
                ('triggered_at', models.DateTimeField(blank=True, null=True)),
                ('is_unique', models.BooleanField()),
                ('is_async', models.BooleanField()),
                ('is_duplicate', models.BooleanField(blank=True, null=True)),
                ('is_on_commit', models.BooleanField()),
                ('input', models.TextField(blank=True, editable=False, null=True)),
                ('task_args',
                 models.JSONField(blank=True, editable=False, encoder=django.core.serializers.json.DjangoJSONEncoder,
                                  null=True)),
                ('task_kwargs',
                 models.JSONField(blank=True, editable=False, encoder=django.core.serializers.json.DjangoJSONEncoder,
                                  null=True)),
                ('estimated_time_of_first_arrival',
                 models.DateTimeField(blank=True, null=True)),
                ('expires_at', models.DateTimeField(blank=True, null=True)),
                ('stale_at', models.DateTimeField(blank=True, null=True)),
                ('state', enumfields.fields.NumEnumField(db_index=True, default=1,
                                                         enum=security.enums.CeleryTaskInvocationLogState)),
            ],
            options={
                'ordering': ('-created_at',),
            },
        ),
        migrations.CreateModel(
            name='CeleryTaskRunLog',
            fields=[
                ('id', models.UUIDField(primary_key=True, serialize=False)),
                ('extra_data',
                 models.JSONField(blank=True, encoder=django.core.serializers.json.DjangoJSONEncoder, null=True)),
                ('slug', models.CharField(blank=True, db_index=True, max_length=255, null=True)),
                ('start', models.DateTimeField(editable=False)),
                ('stop', models.DateTimeField(blank=True, editable=False, null=True)),
                ('time', models.FloatField(blank=True, null=True)),
                ('celery_task_id', models.UUIDField(blank=True, db_index=True, null=True)),
                ('name', models.CharField(db_index=True, max_length=250)),
                ('queue_name', models.CharField(blank=True, max_length=250, null=True)),
                ('input', models.TextField(blank=True, editable=False, null=True)),
                ('task_args',
                 models.JSONField(blank=True, editable=False, encoder=django.core.serializers.json.DjangoJSONEncoder,
                                  null=True)),
                ('task_kwargs',
                 models.JSONField(blank=True, editable=False, encoder=django.core.serializers.json.DjangoJSONEncoder,
                                  null=True)),
                ('state',
                 enumfields.fields.NumEnumField(db_index=True, default=1, enum=security.enums.CeleryTaskRunLogState)),
                ('result',
                 models.JSONField(blank=True, editable=False, encoder=django.core.serializers.json.DjangoJSONEncoder,
                                  null=True)),
                ('error_message', models.TextField(blank=True, null=True)),
                ('output', models.TextField(blank=True, editable=False, null=True)),
                ('retries', models.PositiveSmallIntegerField(default=0)),
                ('estimated_time_of_next_retry',
                 models.DateTimeField(blank=True, null=True)),
            ],
            options={
                'ordering': ('created_at',),
            },
        ),
        migrations.CreateModel(
            name='CommandLog',
            fields=[
                ('id', models.UUIDField(primary_key=True, serialize=False)),
                ('extra_data',
                 models.JSONField(blank=True, encoder=django.core.serializers.json.DjangoJSONEncoder, null=True)),
                ('slug', models.CharField(blank=True, db_index=True, max_length=255, null=True)),
                ('start', models.DateTimeField(editable=False)),
                ('stop', models.DateTimeField(blank=True, editable=False, null=True)),
                ('time', models.FloatField(blank=True, null=True)),
                ('name', models.CharField(db_index=True, editable=False, max_length=250)),
                ('input', models.TextField(editable=False)),
                ('is_executed_from_command_line',
                 models.BooleanField(default=False, editable=False)),
                ('output', models.TextField(blank=True, editable=False, null=True)),
                ('state', enumfields.fields.NumEnumField(db_index=True, default=1, enum=security.enums.CommandState)),
                ('error_message', models.TextField(blank=True, editable=False, null=True)),
            ],
            options={
                'ordering': ('-created_at',),
            },
        ),
        migrations.CreateModel(
            name='InputRequestLog',
            fields=[
                ('id', models.UUIDField(primary_key=True, serialize=False)),
                ('extra_data',
                 models.JSONField(blank=True, encoder=django.core.serializers.json.DjangoJSONEncoder, null=True)),
                ('slug', models.CharField(blank=True, db_index=True, max_length=255, null=True)),
                ('start', models.DateTimeField(editable=False)),
                ('stop', models.DateTimeField(blank=True, editable=False, null=True)),
                ('time', models.FloatField(blank=True, null=True)),
                ('host', models.CharField(db_index=True, max_length=255)),
                ('method', models.SlugField(max_length=7)),
                ('path', models.CharField(blank=True, db_index=True, max_length=2000)),
                ('queries',
                 models.JSONField(blank=True, encoder=django.core.serializers.json.DjangoJSONEncoder, null=True)),
                ('is_secure', models.BooleanField(default=False)),
                ('request_headers',
                 models.JSONField(blank=True, encoder=django.core.serializers.json.DjangoJSONEncoder, null=True)),
                ('request_body', models.TextField(null=True, blank=True)),
                ('response_code', models.PositiveSmallIntegerField(blank=True, null=True)),
                ('response_headers',
                 models.JSONField(blank=True, encoder=django.core.serializers.json.DjangoJSONEncoder, null=True)),
                ('response_body', models.TextField(blank=True, null=True)),
                ('state',
                 enumfields.fields.NumEnumField(default=0, enum=security.enums.RequestLogState)),
                ('error_message', models.TextField(blank=True, null=True)),
                ('user_id', models.TextField(blank=True, db_index=True, null=True)),
                ('ip', models.GenericIPAddressField(db_index=True)),
                ('view_slug',
                 models.CharField(blank=True, db_index=True, max_length=255, null=True)),
            ],
            options={
                'ordering': ('-created_at',),
            },
        ),
        migrations.CreateModel(
            name='OutputRequestLog',
            fields=[
                ('id', models.UUIDField(primary_key=True, serialize=False)),
                ('extra_data',
                 models.JSONField(blank=True, encoder=django.core.serializers.json.DjangoJSONEncoder, null=True)),
                ('slug', models.CharField(blank=True, db_index=True, max_length=255, null=True)),
                ('start', models.DateTimeField(editable=False)),
                ('stop', models.DateTimeField(blank=True, editable=False, null=True)),
                ('time', models.FloatField(blank=True, null=True)),
                ('host', models.CharField(db_index=True, max_length=255)),
                ('method', models.SlugField(max_length=7)),
                ('path', models.CharField(blank=True, db_index=True, max_length=2000)),
                ('queries',
                 models.JSONField(blank=True, encoder=django.core.serializers.json.DjangoJSONEncoder, null=True)),
                ('is_secure', models.BooleanField(default=False)),
                ('request_headers',
                 models.JSONField(blank=True, encoder=django.core.serializers.json.DjangoJSONEncoder, null=True)),
                ('request_body', models.TextField(null=True, blank=True)),
                ('response_code', models.PositiveSmallIntegerField(blank=True, null=True)),
                ('response_headers',
                 models.JSONField(blank=True, encoder=django.core.serializers.json.DjangoJSONEncoder, null=True)),
                ('response_body', models.TextField(blank=True, null=True)),
                ('state',
                 enumfields.fields.NumEnumField(default=0, enum=security.enums.RequestLogState)),
                ('error_message', models.TextField(blank=True, null=True)),
            ],
            options={
                'ordering': ('-created_at',),
            },
        ),
        migrations.CreateModel(
            name='OutputRequestLogGenericManyToManyRelation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False)),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True)),
                ('changed_at', models.DateTimeField(auto_now=True, db_index=True)),
                ('object_ct_id',
                 models.PositiveSmallIntegerField(db_index=True)),
                ('object_id', models.TextField(db_index=True)),
                ('output_request_log',
                 models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='_related_objects',
                                   related_query_name='related_objects', to='security_backends_sql.outputrequestlog')),
            ],
            options={
                'db_tablespace': '',
                'unique_together': {('output_request_log', 'object_ct_id', 'object_id')},
            },
        ),
        migrations.CreateModel(
            name='InputRequestLogGenericManyToManyRelation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False)),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True)),
                ('changed_at', models.DateTimeField(auto_now=True, db_index=True)),
                ('object_ct_id',
                 models.PositiveSmallIntegerField(db_index=True)),
                ('object_id', models.TextField(db_index=True)),
                ('input_request_log',
                 models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='_related_objects',
                                   related_query_name='related_objects', to='security_backends_sql.inputrequestlog')),
            ],
            options={
                'db_tablespace': '',
                'unique_together': {('input_request_log', 'object_ct_id', 'object_id')},
            },
        ),
        migrations.CreateModel(
            name='CommandLogGenericManyToManyRelation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False)),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True)),
                ('changed_at', models.DateTimeField(auto_now=True, db_index=True)),
                ('object_ct_id',
                 models.PositiveSmallIntegerField(db_index=True)),
                ('object_id', models.TextField(db_index=True)),
                ('command_log',
                 models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='_related_objects',
                                   related_query_name='related_objects', to='security_backends_sql.commandlog')),
            ],
            options={
                'db_tablespace': '',
                'unique_together': {('command_log', 'object_ct_id', 'object_id')},
            },
        ),
        migrations.CreateModel(
            name='CeleryTaskRunLogGenericManyToManyRelation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False)),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True)),
                ('changed_at', models.DateTimeField(auto_now=True, db_index=True)),
                ('object_ct_id',
                 models.PositiveSmallIntegerField(db_index=True)),
                ('object_id', models.TextField(db_index=True)),
                ('celery_task_run_log',
                 models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='_related_objects',
                                   related_query_name='related_objects', to='security_backends_sql.celerytaskrunlog')),
            ],
            options={
                'db_tablespace': '',
                'unique_together': {('celery_task_run_log', 'object_ct_id', 'object_id')},
            },
        ),
        migrations.CreateModel(
            name='CeleryTaskInvocationLogGenericManyToManyRelation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False)),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True)),
                ('changed_at', models.DateTimeField(auto_now=True, db_index=True)),
                ('object_ct_id',
                 models.PositiveSmallIntegerField(db_index=True)),
                ('object_id', models.TextField(db_index=True)),
                ('celery_task_invocation_log',
                 models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='_related_objects',
                                   related_query_name='related_objects',
                                   to='security_backends_sql.celerytaskinvocationlog')),
            ],
            options={
                'db_tablespace': '',
                'unique_together': {('celery_task_invocation_log', 'object_ct_id', 'object_id')},
            },
        ),
    ]



import traceback

from django import http
from django.conf import settings as django_settings
from django.urls import is_valid_path, get_callable, get_resolver, set_urlconf, Resolver404

from ipware.ip import get_client_ip

from security.logging.requests.logger import InputRequestLogger
from security.throttling.exception import ThrottlingException

from .config import settings
from .utils import get_throttling_validators


def get_view_from_request_or_none(request):
    try:
        if hasattr(request, 'urlconf'):
            urlconf = request.urlconf
            set_urlconf(urlconf)
            resolver = get_resolver(urlconf)
        else:
            resolver = get_resolver()
        return resolver.resolve(request.path_info)[0]
    except Resolver404:
        return None


class LogMiddleware:

    response_redirect_class = http.HttpResponsePermanentRedirect

    def __init__(self, get_response=None):
        self.base_get_response = get_response
        super().__init__()

    def __call__(self, request):
        view = get_view_from_request_or_none(request)
        if (get_client_ip(request)[0] not in settings.LOG_REQUEST_IGNORE_IP
                and request.path not in settings.LOG_REQUEST_IGNORE_URL_PATHS
                and not getattr(view, 'log_exempt', False)):

            with InputRequestLogger(hide_request_body=getattr(view, 'hide_request_body', False)) as logger:
                request.input_request_logger = logger
                return self.get_response(request, logger)
        else:
            return self.get_response(request)

    def get_response(self, request, input_request_logger=None):
        response = self.process_request(request, input_request_logger)
        response = response or self.base_get_response(request)
        return self.process_response(request, response, input_request_logger)

    def process_request(self, request, input_request_logger=None):
        if input_request_logger:
            input_request_logger.log_request(request)

        # Return a redirect if necessary
        if self.should_redirect_with_slash(request):
            return self.response_redirect_class(self.get_full_path_with_slash(request))

    def get_full_path_with_slash(self, request):
        """
        Return the full path of the request with a trailing slash appended.

        Raise a RuntimeError if settings.DEBUG is True and request.method is
        POST, PUT, or PATCH.
        """
        new_path = request.get_full_path(force_append_slash=True)
        if django_settings.DEBUG and request.method in {'POST', 'PUT', 'PATCH'}:
            raise RuntimeError(
                "You called this URL via %(method)s, but the URL doesn't end "
                "in a slash and you have SECURITY_APPEND_SLASH set. Django can't "
                "redirect to the slash URL while maintaining %(method)s data. "
                "Change your form to point to %(url)s (note the trailing "
                "slash), or set SECURITY_APPEND_SLASH=False in your Django settings." % {
                    'method': request.method,
                    'url': request.get_host() + new_path,
                }
            )
        return new_path

    def should_redirect_with_slash(self, request):
        """
        Return True if settings.APPEND_SLASH is True and appending a slash to
        the request path turns an invalid path into a valid one.
        """
        if settings.APPEND_SLASH and not request.get_full_path().endswith('/'):
            urlconf = getattr(request, 'urlconf', None)
            return (
                not is_valid_path(request.path_info, urlconf) and
                is_valid_path('{}/'.format(request.path_info), urlconf)
            )
        return False

    def _render_throttling(self, request, exception):
        return get_callable(settings.THROTTLING_FAILURE_VIEW)(request, exception)

    def process_view(self, request, callback, callback_args, callback_kwargs):
        throttling_validators = getattr(callback, 'throttling_validators', None)
        if throttling_validators is None:
            throttling_validators = get_throttling_validators('default_validators')

        try:
            for validator in throttling_validators:
                validator.validate(request)
        except ThrottlingException as exception:
            return self.process_exception(request, exception)

    def process_response(self, request, response, input_request_logger=None):
        if input_request_logger:
            input_request_logger.log_response(request, response)
        return response

    def process_exception(self, request, exception):
        input_request_logger = getattr(request, 'input_request_logger', None)
        if input_request_logger:
            input_request_logger.log_exception(traceback.format_exc())
            if isinstance(exception, ThrottlingException):
                return self._render_throttling(request, exception)



import traceback

from requests import *  # noqa: F403, F401
from requests import Session, Request

from security.logging.requests.logger import OutputRequestLogger


class SecuritySession(Session):

    def __init__(self, slug=None, related_objects=None):
        super().__init__()
        self.slug = slug
        self.related_objects = [] if related_objects is None else related_objects

    def request(self, method, url, params=None, data=None, headers=None, cookies=None, files=None, auth=None,
                timeout=None, allow_redirects=True, proxies=None, hooks=None, stream=None, verify=None, cert=None,
                json=None, slug=None, related_objects=None):

        related_objects = [] if related_objects is None else related_objects

        # Create the Request.
        req = Request(
            method=method.upper(), url=url, headers=headers, files=files, data=data or {}, json=json,
            params=params or {}, auth=auth, cookies=cookies, hooks=hooks,
        )
        prep = self.prepare_request(req)
        proxies = proxies or {}
        request_settings = self.merge_environment_settings(
            prep.url, proxies, stream, verify, cert
        )
        # Send the request.
        send_kwargs = {
            'timeout': timeout,
            'allow_redirects': allow_redirects,
        }
        send_kwargs.update(request_settings)

        with OutputRequestLogger(related_objects=related_objects or self.related_objects,
                                 slug=slug or self.slug) as output_request_logger:
            output_request_logger.log_request(prep)
            try:
                response = self.send(prep, **send_kwargs)
                output_request_logger.log_response(response)
                return response
            except Exception:
                output_request_logger.log_exception(traceback.format_exc())
                raise


def request(method, url, **kwargs):
    with SecuritySession() as session:
        return session.request(method=method, url=url, **kwargs)


def get(url, params=None, **kwargs):
    """Sends a GET request.
    :param url: URL for the new :class:`Request` object.
    :param params: (optional) Dictionary or bytes to be sent in the query string for the :class:`Request`.
    :param kwargs: Optional arguments that ``request`` takes.
    :return: :class:`Response <Response>` object
    :rtype: requests.Response
    """

    kwargs.setdefault('allow_redirects', True)
    return request('get', url, params=params, **kwargs)


def options(url, **kwargs):
    """Sends a OPTIONS request.
    :param url: URL for the new :class:`Request` object.
    :param kwargs: Optional arguments that ``request`` takes.
    :return: :class:`Response <Response>` object
    :rtype: requests.Response
    """

    kwargs.setdefault('allow_redirects', True)
    return request('options', url, **kwargs)


def head(url, **kwargs):
    """Sends a HEAD request.
    :param url: URL for the new :class:`Request` object.
    :param kwargs: Optional arguments that ``request`` takes.
    :return: :class:`Response <Response>` object
    :rtype: requests.Response
    """

    kwargs.setdefault('allow_redirects', False)
    return request('head', url, **kwargs)


def post(url, data=None, json=None, **kwargs):
    """Sends a POST request.
    :param url: URL for the new :class:`Request` object.
    :param data: (optional) Dictionary, bytes, or file-like object to send in the body of the :class:`Request`.
    :param json: (optional) json data to send in the body of the :class:`Request`.
    :param kwargs: Optional arguments that ``request`` takes.
    :return: :class:`Response <Response>` object
    :rtype: requests.Response
    """

    return request('post', url, data=data, json=json, **kwargs)


def put(url, data=None, **kwargs):
    """Sends a PUT request.
    :param url: URL for the new :class:`Request` object.
    :param data: (optional) Dictionary, bytes, or file-like object to send in the body of the :class:`Request`.
    :param kwargs: Optional arguments that ``request`` takes.
    :return: :class:`Response <Response>` object
    :rtype: requests.Response
    """

    return request('put', url, data=data, **kwargs)


def patch(url, data=None, **kwargs):
    """Sends a PATCH request.
    :param url: URL for the new :class:`Request` object.
    :param data: (optional) Dictionary, bytes, or file-like object to send in the body of the :class:`Request`.
    :param kwargs: Optional arguments that ``request`` takes.
    :return: :class:`Response <Response>` object
    :rtype: requests.Response
    """

    return request('patch', url, data=data, **kwargs)


def delete(url, **kwargs):
    """Sends a DELETE request.
    :param url: URL for the new :class:`Request` object.
    :param kwargs: Optional arguments that ``request`` takes.
    :return: :class:`Response <Response>` object
    :rtype: requests.Response
    """

    return request('delete', url, **kwargs)



from django.contrib.auth.models import User
from django.core.management.base import BaseCommand

from reversion.revisions import create_revision


class Command(BaseCommand):

    @create_revision()
    def handle(self, **options):
        User.objects._create_user('test', 'test@localhost', 'test', is_staff=True, is_superuser=True)



from security.config import settings

from . import db_helpers


class MirrorSecurityLoggerRouter:
    """
    A router controlling tha all that read and write database operations will be performed via mirror without atomic.
    """

    route_app_labels = [
        'security_backends_sql'
    ]

    def _db_for_read_or_write(self, model):
        if not db_helpers.is_running_migration and model._meta.app_label in self.route_app_labels:
            return settings.LOG_DB_NAME
        return None

    def db_for_read(self, model, **hints):
        return self._db_for_read_or_write(model)

    def db_for_write(self, model, **hints):
        return self._db_for_read_or_write(model)

    def allow_relation(self, obj1, obj2, **hints):
        return True

    def allow_migrate(self, db, app_label, model_name=None, **hints):
        return True


class MultipleDBSecurityLoggerRouter:
    """
    A router controlling that all read and write database operations for apps specified in route_app_labels
    will be performed via mirror without atomic.
    """

    route_app_labels = [
        'security_backends_sql'
    ]

    def _db_for_read_or_write(self, model):
        if model._meta.app_label in self.route_app_labels:
            return settings.LOG_DB_NAME
        return None

    def db_for_read(self, model, **hints):
        return self._db_for_read_or_write(model)

    def db_for_write(self, model, **hints):
        return self._db_for_read_or_write(model)

    def allow_relation(self, obj1, obj2, **hints):
        return None

    def allow_migrate(self, db, app_label, model_name=None, **hints):
        if app_label in self.route_app_labels:
            return db == settings.LOG_DB_NAME
        else:
            return db != settings.LOG_DB_NAME



import json

from ansi2html import Ansi2HTMLConverter

from django.contrib.contenttypes.models import ContentType
from django.template.defaultfilters import truncatechars
from django.utils.translation import ugettext_lazy as _
from django.utils.html import mark_safe

from chamber.shortcuts import get_object_or_none

from pyston.utils.decorators import filter_by, order_by

from is_core.generic_views.detail_views import DjangoReadonlyDetailView
from is_core.utils import display_code, render_model_object_with_link, render_model_objects_with_link
from is_core.utils.decorators import short_description


field_labels = {
    'id': _('ID'),
    'release': _('release'),
    'slug': _('slug'),
    'view_slug': _('view slug'),
    'host': _('host'),
    'path': _('path'),
    'state': _('state'),
    'start': _('start'),
    'stop': _('stop'),
    'time': _('time'),
    'method': _('method'),
    'response_code': _('response code'),
    'ip': _('IP'),
    'is_secure': _('HTTPS connection'),
    'short_path': _('pah'),
    'short_queries': _('queries'),
    'short_request_headers': _('request headers'),
    'short_request_body': _('request body'),
    'short_response_headers': _('response headers'),
    'short_response_body': _('response body'),
    'name': _('name'),
    'input': _('input'),
    'is_executed_from_command_line': _('is executed from command line'),
    'celery_task_id': _('celery task ID'),
    'queue_name': _('queue name'),
    'applied_at': _('applied at'),
    'triggered_at': _('triggered at'),
    'is_unique': _('is unique'),
    'is_async': _('is async'),
    'is_duplicate': _('is duplicate'),
    'is_on_commit': _('is on commit'),
    'estimated_time_of_first_arrival': _('estimated time of first arrival'),
    'expires_at': _('expires at'),
    'stale_at': _('stale at'),
    'retries': _('retries'),
    'estimated_time_of_next_retry': _('estimated time of next retry'),
}


def display_json(value, indent=4):
    dict_value = json.loads(value) if isinstance(value, str) else value
    return json.dumps(dict_value, indent=indent, ensure_ascii=False)


class ChildLogTableViewMixin:

    fields = (
        'id', 'start', 'stop'
    )

    def _get_log_key(self, log):
        raise NotImplementedError

    def _get_list_filter(self):
        return {
            'filter': {
                'parent_log': self._get_log_key(self.parent_instance)
            }
        }


class RelatedLogInlineTableViewMixin:

    def _get_list_filter(self):
        return {
            'filter': {
                'display_related_objects__in': self.parent_view.get_related_obj_keys(self.parent_instance)
            }
        }


class CeleryTaskInvocationLogInlineTableViewMixin:

    fields = (
        'id', 'start', 'stop', 'time', 'state',
    )

    def _get_list_filter(self):
        return {
            'filter': {
                'celery_task_id': self.parent_instance.celery_task_id
            }
        }


class CeleryTaskRunLogInlineTableViewMixin:

    fields = (
        'start', 'stop', 'waiting_time', 'time', 'state', 'retries'
    )

    def _get_list_filter(self):
        return {
            'filter': {
                'celery_task_id': self.parent_instance.celery_task_id
            }
        }


class LogCoreMixin:

    field_labels = field_labels

    default_ordering = ('-start',)
    rest_extra_filter_fields = ('parent_log',)

    can_create = False
    can_read = True
    can_update = False
    can_delete = False

    output_request_inline_table_view = None
    command_inline_table_view = None
    celery_task_invocation_inline_table_view = None

    display_related_objects_filter = None

    def __new__(cls, *args, **kwargs):
        if cls.display_related_objects_filter is not None:
            cls.display_related_objects.filter = cls.display_related_objects_filter
            cls.rest_extra_filter_fields = list(cls.rest_extra_filter_fields or ()) + [
                'display_related_objects',
            ]
        return super().__new__(cls)

    @short_description(_('error message'))
    @filter_by('error_message')
    def error_message_code(self, obj):
        return display_code(obj.error_message) if obj else None

    @short_description(_('source'))
    def display_source(self, obj, request):
        parent_log_obj = self._get_parent_log_obj(obj)
        return render_model_object_with_link(request, parent_log_obj) if parent_log_obj else None

    def _get_parent_log_obj(self, obj):
        raise NotImplementedError

    def _get_related_objects(self, obj):
        raise NotImplementedError

    @short_description(_('related objects'))
    def display_related_objects(self, obj, request):
        return render_model_objects_with_link(request, self._get_related_objects(obj))

    @short_description(_('revisions'))
    def revisions(self, obj, request):
        try:
            from reversion.models import get_revision_or_none
        except ImportError:
            def get_revision_or_none(id):
                try:
                    from reversion.models import Revision
                    return get_object_or_none(Revision, pk=id)
                except ImportError:
                    return None

        if 'reversion' in obj.extra_data:
            reversion_data = obj.extra_data['reversion']
            revisions_display = []
            for revision_data in reversion_data['revisions']:
                revision_obj = get_revision_or_none(revision_data['id'])
                revision_with_link = render_model_object_with_link(request, revision_obj) if revision_obj else None
                if revision_with_link:
                    revisions_display.append(revision_with_link)
            if len(reversion_data['revisions']) != reversion_data['total_count']:
                revisions_display.append('…')
            return revisions_display
        else:
            return None


class RequestLogCoreMixin(LogCoreMixin):

    @filter_by('path')
    @order_by('path')
    @short_description(_('path'))
    def short_path(self, obj):
        return truncatechars(obj.path, 50)

    @filter_by('request_body')
    @order_by('request_body')
    @short_description(_('request body'))
    def short_request_body(self, obj):
        return truncatechars(obj.request_body, 50)

    @filter_by('response_body')
    @order_by('response_body')
    @short_description(_('response body'))
    def short_response_body(self, obj):
        return truncatechars(obj.response_body, 50) if obj.response_body is not None else None

    @filter_by('queries')
    @order_by('queries')
    @short_description(_('queries'))
    def short_queries(self, obj):
        return truncatechars(obj.queries, 50)

    @filter_by('request_headers')
    @order_by('request_headers')
    @short_description(_('request headers'))
    def short_request_headers(self, obj):
        return truncatechars(obj.request_headers, 50)

    @filter_by('request_headers')
    @order_by('request_headers')
    @short_description(_('request headers'))
    def short_response_headers(self, obj):
        return truncatechars(obj.response_headers, 50)

    @short_description(_('request body'))
    def request_body_code(self, obj):
        return display_code(obj.request_body) if obj and obj.request_body else None

    @short_description(_('response body'))
    def response_body_code(self, obj):
        return display_code(obj.response_body) if obj and obj.response_body else None

    @short_description(_('queries'))
    def queries_code(self, obj):
        return display_code(display_json(obj.queries)) if obj and obj.queries else None

    @short_description(_('request headers'))
    def request_headers_code(self, obj):
        return display_code(display_json(obj.request_headers)) if obj and obj.request_headers else None

    @short_description(_('response headers'))
    def response_headers_code(self, obj):
        return display_code(display_json(obj.response_headers)) if obj and obj.response_headers else None


class InputRequestLogCoreMixin(RequestLogCoreMixin):

    menu_group = 'inputrequestlog'

    list_fields = (
        'id', 'start', 'stop', 'release', 'slug', 'view_slug', 'host', 'short_path', 'state', 'time', 'method',
        'short_queries', 'response_code', 'short_request_headers', 'short_request_body', 'short_response_headers',
        'short_response_body', 'user', 'ip',
    )

    verbose_name = _('input request log')
    verbose_name_plural = _('input request logs')

    def get_fieldsets(self, request, obj=None):
        return (
            (None, {
                'fields': (
                    'id', 'release', 'slug', 'view_slug', 'host', 'path', 'state', 'start', 'stop', 'time', 'method',
                    'is_secure', 'user', 'ip'
                ),
            }),
            (_('request'), {'fields': ('queries_code', 'request_headers_code', 'request_body_code')}),
            (_('response'), {'fields': ('response_code', 'state', 'response_headers_code', 'response_body_code')}),
            (_('output'), {
                'fields': (
                    'error_message_code',
                ),
            }),
            (_('relations'), {
                'fieldsets': (
                    (None, {'fields': (
                        'revisions', 'display_source', 'display_related_objects'
                    )}),
                    (_('output requests'), {'inline_view': self.output_request_inline_table_view}),
                    (_('commands'), {'inline_view': self.command_inline_table_view}),
                    (_('celery invocations'), {'inline_view': self.celery_task_invocation_inline_table_view}),
                ),
            }),
            (None, {'fields': ('debug_toolbar',)})
        )

    @short_description('')
    def debug_toolbar(self, obj):
        return (
            mark_safe(obj.extra_data.get('debug_toolbar'))
            if obj.extra_data and 'debug_toolbar' in obj.extra_data else ''
        )


class OutputRequestLogCoreMixin(RequestLogCoreMixin):

    menu_group = 'outputrequestlog'

    list_fields = (
        'id', 'start', 'stop', 'release', 'slug', 'host', 'short_path', 'state', 'time', 'method', 'short_queries',
        'response_code', 'short_request_headers', 'short_request_body', 'short_response_headers', 'short_response_body'
    )

    fieldsets = (
        (None, {
            'fields': (
                'id', 'release', 'slug', 'host', 'path', 'state', 'start', 'stop', 'time', 'method', 'is_secure',
            ),
        }),
        (_('request'), {'fields': ('queries_code', 'request_headers_code', 'request_body_code')}),
        (_('response'), {'fields': ('response_code', 'state', 'response_headers_code', 'response_body_code')}),
        (_('output'), {
            'fields': (
                'error_message_code',
            ),
        }),
        (_('relations'), {
            'fields': (
                'display_source', 'display_related_objects'
            ),
        }),
    )

    verbose_name = _('output request log')
    verbose_name_plural = _('output request logs')


class OutputLogCoreMixin:

    @short_description(_('output'))
    def output_html(self, obj=None):
        if obj and obj.output is not None:
            return display_code(mark_safe(Ansi2HTMLConverter().convert(obj.output, full=False)))
        return None


class CommandLogCoreMixin(OutputLogCoreMixin, LogCoreMixin):

    menu_group = 'commandlog'

    list_fields = (
        'id', 'start', 'stop', 'release', 'slug', 'name', 'state', 'time'
    )

    verbose_name = _('command log')
    verbose_name_plural = _('command logs')

    def get_fieldsets(self, request, obj=None):
        return (
            (None, {
                'fields': (
                    'id', 'release', 'slug', 'name', 'state', 'start', 'stop', 'time', 'input',
                    'is_executed_from_command_line',
                ),
            }),
            (_('output'), {
                'fields': (
                    'error_message_code', 'output_html',
                ),
            }),
            (_('relations'), {
                'fieldsets': (
                    (None, {'fields': (
                        'revisions', 'display_source', 'display_related_objects'
                    )}),
                    (_('output requests'), {'inline_view': self.output_request_inline_table_view}),
                    (_('commands'), {'inline_view': self.command_inline_table_view}),
                    (_('celery invocations'), {'inline_view': self.celery_task_invocation_inline_table_view}),
                ),
            }),
        )


class CeleryCoreMixin(LogCoreMixin):

    @short_description(_('task args'))
    def task_args_code(self, obj):
        return display_code(display_json(obj.task_args)) if obj and obj.task_args else None

    @short_description(_('task kwargs'))
    def task_kwargs_code(self, obj):
        return display_code(display_json(obj.task_kwargs)) if obj and obj.task_kwargs else None

    @filter_by('input')
    @order_by('input')
    @short_description(_('input'))
    def short_input(self, obj):
        return truncatechars(obj.input, 50)


class CeleryTaskRunLogCoreMixin(OutputLogCoreMixin, CeleryCoreMixin, LogCoreMixin):

    menu_group = 'celerytaskrunlog'

    rest_fields = list_fields = (
        'id', 'start', 'stop', 'release', 'slug', 'celery_task_id', 'name', 'state', 'waiting_time', 'time',
        'short_input', 'queue_name'
    )

    verbose_name = _('celery task run log')
    verbose_name_plural = _('celery task run logs')

    def get_fieldsets(self, request, obj=None):
        return (
            (None, {
                'fields': (
                    'id', 'release', 'celery_task_id', 'slug', 'name', 'state', 'start', 'stop', 'waiting_time', 'time',
                    'input', 'task_args_code', 'task_kwargs_code', 'retries', 'estimated_time_of_next_retry',
                    'queue_name'
                ),
            }),
            (_('output'), {
                'fields': (
                    'error_message_code', 'output_html', 'result_code'
                ),
            }),
            (_('relations'), {
                'fieldsets': (
                    (None, {'fields': (
                        'revisions', 'display_related_objects',
                    )}),
                    (_('output requests'), {'inline_view': self.output_request_inline_table_view}),
                    (_('commands'), {'inline_view': self.command_inline_table_view}),
                    (_('celery invocations'), {'inline_view': self.celery_task_invocation_inline_table_view}),
                ),
            }),
        )

    @short_description(_('result'))
    def result_code(self, obj):
        return display_code(display_json(obj.result)) if obj and obj.result else None


class CeleryTaskInvocationLogCoreMixin(CeleryCoreMixin, LogCoreMixin):

    menu_group = 'celerytaskinvocationlog'

    rest_fields = list_fields = (
        'id', 'start', 'stop', 'release', 'slug', 'celery_task_id', 'name', 'state_with_duplicate', 'time',
        'short_input', 'queue_name'
    )

    celery_task_run_inline_table_view = None

    verbose_name = _('celery task invocation log')
    verbose_name_plural = _('celery task invocation logs')

    @filter_by('state')
    @order_by('state')
    @short_description(_('state'))
    def state_with_duplicate(self, obj):
        return '{} ({})'.format(obj.state.label, _('duplicate')) if obj.is_duplicate else obj.state.label

    def get_fieldsets(self, request, obj=None):
        return (
            (None, {
                'fields': (
                    'id', 'release', 'celery_task_id', 'slug', 'name', 'state_with_duplicate', 'start', 'stop', 'time',
                    'input', 'task_args_code', 'task_kwargs_code', 'applied_at', 'triggered_at', 'queue_name',
                    'is_unique', 'is_async', 'is_duplicate', 'is_on_commit', 'estimated_time_of_first_arrival',
                    'expires_at', 'stale_at'
                ),
            }),
            (_('celery task runs'), {'inline_view': self.celery_task_run_inline_table_view}),
            (_('relations'), {
                'fields': (
                    'display_source', 'display_related_objects'
                ),
            }),
        )


class BaseRelatedLogsView(DjangoReadonlyDetailView):

    input_request_inline_view = None
    output_request_inline_view = None
    command_inline_view = None
    celery_task_run_inline_view = None
    celery_task_invocation_inline_view = None

    show_input_request = show_output_request = show_command = show_celery_task_run = show_celery_task_invocation = True

    title = _('related logs')

    inline_view_labels = {
        'input_request': _('input request logs'),
        'output_request': _('output request logs'),
        'command': _('command logs'),
        'celery_task_run': _('celery task run logs'),
        'celery_task_invocation': _('celery task invocation logs'),
    }

    def get_fieldsets(self):
        fieldsets = []
        for view_name, label in self.inline_view_labels.items():
            show_inline_view = getattr(self, f'show_{view_name}')
            inline_view_class = getattr(self, f'{view_name}_inline_view')
            if show_inline_view and inline_view_class:
                fieldsets.append((
                    label, {'inline_view': inline_view_class}
                ))
        return fieldsets

    def _get_obj_key(self, obj):
        return f'{ContentType.objects.get_for_model(obj).pk}|{obj.pk}'

    def _get_related_objs(self, obj):
        return [obj]

    def get_related_obj_keys(self, obj):
        return [self._get_obj_key(o) for o in self._get_related_objs(obj)]



from django.utils.timezone import now

from security.enums import LoggerName
from security.logging.common import SecurityLogger
from security.backends.signals import (
    celery_task_invocation_started, celery_task_invocation_triggered, celery_task_invocation_ignored,
    celery_task_invocation_timeout, celery_task_invocation_expired, celery_task_run_started, celery_task_run_succeeded,
    celery_task_run_failed, celery_task_run_retried, celery_task_run_output_updated
)


class CeleryInvocationLogger(SecurityLogger):

    name = LoggerName.CELERY_TASK_INVOCATION
    store = False

    def log_started(self, name, queue_name, task_args, task_kwargs, applied_at,
                    is_async, is_unique, is_on_commit):
        task_input = []
        if task_args:
            task_input += [str(v) for v in task_args]
        if task_kwargs:
            task_input += ['{}={}'.format(k, v) for k, v in task_kwargs.items()]
        self.data.update(dict(
            name=name,
            queue_name=queue_name,
            input=', '.join(task_input),
            task_args=list(task_args),
            task_kwargs=task_kwargs,
            applied_at=applied_at,
            is_async=is_async,
            is_unique=is_unique,
            is_on_commit=is_on_commit,
            start=now()
        ))
        celery_task_invocation_started.send(
            sender=CeleryInvocationLogger,
            logger=self,
        )

    def log_triggered(self, triggered_at, stale_at, estimated_time_of_first_arrival,
                      expires_at, celery_task_id):
        self.data.update(dict(
            triggered_at=triggered_at,
            stale_at=stale_at,
            estimated_time_of_first_arrival=estimated_time_of_first_arrival,
            expires_at=expires_at,
            celery_task_id=celery_task_id,
            is_duplicate=False
        ))
        celery_task_invocation_triggered.send(
            sender=CeleryInvocationLogger,
            logger=self,
        )

    def log_unique(self, triggered_at, stale_at, estimated_time_of_first_arrival,
                   expires_at, celery_task_id):
        self.data.update(dict(
            triggered_at=triggered_at,
            stale_at=stale_at,
            estimated_time_of_first_arrival=estimated_time_of_first_arrival,
            expires_at=expires_at,
            celery_task_id=celery_task_id,
            is_duplicate=True
        ))
        celery_task_invocation_triggered.send(
            sender=CeleryInvocationLogger,
            logger=self,
        )

    def log_ignored(self):
        self.data.update(dict(
            stop=now()
        ))
        celery_task_invocation_ignored.send(
            sender=CeleryInvocationLogger,
            logger=self,
        )

    def log_timeout(self):
        self.data.update(dict(
            stop=now()
        ))
        celery_task_invocation_timeout.send(
            sender=CeleryInvocationLogger,
            logger=self,
        )

    def log_expired(self, name):
        self.data.update(dict(
            stop=now()
        ))
        celery_task_invocation_expired.send(
            sender=CeleryInvocationLogger,
            name=name,
            logger=self,
        )


class CeleryTaskRunLogger(SecurityLogger):

    name = LoggerName.CELERY_TASK_RUN

    def log_started(self, name, queue_name, task_args, task_kwargs, celery_task_id, retries, trigger_time):
        task_input = []
        if task_args:
            task_input += [str(v) for v in task_args]
        if task_kwargs:
            task_input += ['{}={}'.format(k, v) for k, v in task_kwargs.items()]
        start = now()
        self.data.update(dict(
            name=name,
            celery_task_id=celery_task_id,
            queue_name=queue_name,
            input=', '.join(task_input),
            task_args=list(task_args),
            task_kwargs=task_kwargs,
            start=start,
            retries=retries,
            waiting_time=(start - trigger_time).total_seconds(),
        ))
        celery_task_run_started.send(
            sender=CeleryTaskRunLogger,
            logger=self,
        )

    def log_succeeded(self, result):
        self.data.update(dict(
            result=result,
            stop=now(),
        ))
        celery_task_run_succeeded.send(
            sender=CeleryTaskRunLogger,
            logger=self,
        )

    def log_failed(self, ex_tb):
        self.data.update(dict(
            error_message=ex_tb,
            stop=now(),
        ))
        celery_task_run_failed.send(
            sender=CeleryTaskRunLogger,
            logger=self,
        )

    def log_retried(self, ex_tb, estimated_time_of_next_retry):
        self.data.update(dict(
            error_message=ex_tb,
            stop=now(),
            estimated_time_of_next_retry=estimated_time_of_next_retry
        ))
        celery_task_run_retried.send(
            sender=CeleryTaskRunLogger,
            logger=self,
        )

    def log_output_updated(self, output):
        self.data['output'] = output
        celery_task_run_output_updated.send(sender=CeleryTaskRunLogger, logger=self)



import logging

from django.core.exceptions import ImproperlyConfigured

from security.backends.signals import (
    input_request_started, input_request_finished, input_request_error,
    output_request_started, output_request_finished, output_request_error,
    command_started, command_output_updated, command_finished, command_error,
    celery_task_invocation_started, celery_task_invocation_triggered, celery_task_invocation_ignored,
    celery_task_invocation_timeout, celery_task_invocation_expired,
    celery_task_run_started, celery_task_run_succeeded, celery_task_run_failed, celery_task_run_retried,
    celery_task_run_output_updated
)
from security.config import settings

from .app import SecurityBackend


logging_logger = logging.getLogger(__name__)


class BaseBackendWriter:

    CAPTURED_SIGNALS = {
        'input_request_started': input_request_started,
        'input_request_finished': input_request_finished,
        'input_request_error': input_request_error,

        'output_request_started': output_request_started,
        'output_request_finished': output_request_finished,
        'output_request_error': output_request_error,

        'command_started': command_started,
        'command_output_updated': command_output_updated,
        'command_finished': command_finished,
        'command_error': command_error,

        'celery_task_invocation_started': celery_task_invocation_started,
        'celery_task_invocation_triggered': celery_task_invocation_triggered,
        'celery_task_invocation_ignored': celery_task_invocation_ignored,
        'celery_task_invocation_timeout': celery_task_invocation_timeout,
        'celery_task_invocation_expired': celery_task_invocation_expired,

        'celery_task_run_started': celery_task_run_started,
        'celery_task_run_succeeded': celery_task_run_succeeded,
        'celery_task_run_failed': celery_task_run_failed,
        'celery_task_run_retried': celery_task_run_retried,
        'celery_task_run_output_updated': celery_task_run_output_updated,
    }

    def __init__(self, name):
        self._name = name
        self._init_signals()

    def _call_receiver_method(self, signal_name, logger):
        try:
            getattr(self, signal_name)(logger)
        except:  # noqa: E722
            logging_logger.error(f'Cannot write log {signal_name} for writer {self._name}', exc_info=True)

    def _get_receiver(self, signal_name):
        def _log_receiver(sender, logger, signal, **kwargs):
            if settings.BACKEND_WRITERS is None or self._name in settings.BACKEND_WRITERS:
                self._call_receiver_method(signal_name, logger)
        return _log_receiver

    def _init_signals(self):
        for signal_name, signal in self.CAPTURED_SIGNALS.items():
            signal.connect(self._get_receiver(signal_name), weak=False)

    def input_request_started(self, logger):
        raise NotImplementedError

    def input_request_finished(self, logger):
        raise NotImplementedError

    def input_request_error(self, logger):
        raise NotImplementedError

    def output_request_started(self, logger):
        raise NotImplementedError

    def output_request_finished(self, logger):
        raise NotImplementedError

    def output_request_error(self, logger):
        raise NotImplementedError

    def command_started(self, logger):
        raise NotImplementedError

    def command_output_updated(self, logger):
        raise NotImplementedError

    def command_finished(self, logger):
        raise NotImplementedError

    def command_error(self, logger):
        raise NotImplementedError

    def celery_task_invocation_started(self, logger):
        raise NotImplementedError

    def celery_task_invocation_triggered(self, logger):
        raise NotImplementedError

    def celery_task_invocation_ignored(self, logger):
        raise NotImplementedError

    def celery_task_invocation_timeout(self, logger):
        raise NotImplementedError

    def celery_task_invocation_expired(self, logger):
        raise NotImplementedError

    def celery_task_run_started(self, logger):
        raise NotImplementedError

    def celery_task_run_succeeded(self, logger):
        raise NotImplementedError

    def celery_task_run_failed(self, logger):
        raise NotImplementedError

    def celery_task_run_retried(self, logger):
        raise NotImplementedError

    def celery_task_run_output_updated(self, logger):
        raise NotImplementedError

    def set_stale_celery_task_log_state(self):
        pass

    def clean_logs(self, type, timestamp, backup_path, stdout):
        pass


def get_writer_backends():
    if not SecurityBackend.registered_readers:
        raise ImproperlyConfigured('No registered backend reader was set')
    if settings.BACKEND_WRITERS is not None:
        registered_writers = []
        for backend_name in settings.BACKEND_WRITERS:
            if backend_name not in SecurityBackend.registered_writers:
                raise ImproperlyConfigured(f'Backend writer "{backend_name}" is not registered')
            registered_writers.append(SecurityBackend.registered_writers[backend_name])
        return registered_writers
    else:
        return SecurityBackend.registered_writers.values()


def set_stale_celery_task_log_state():
    for writer in get_writer_backends():
        writer.set_stale_celery_task_log_state()


def clean_logs(type, timestamp, backup_path, stdout):
    for writer in get_writer_backends():
        writer.clean_logs(type, timestamp, backup_path, stdout)



import sys
from distutils.version import StrictVersion

import django
from django.core.management import CommandError, CommandParser
from django.core.management import call_command as call_command_original
from django.core.management import execute_from_command_line as execute_from_command_line_original
from django.core.management import handle_default_options
from django.utils.version import get_main_version


def execute_from_command_line(argv=None):
    from security.command import CommandExecutor

    def execute_from_command_line_with_stdout(argv=None, stdout=None, stderr=None):
        try:
            from django.core.management.base import BaseCommand

            command_stdout = stdout
            command_stderr = stderr

            def command_init_patch(self, stdout=None, stderr=None, no_color=False, force_color=False):
                stdout = command_stdout if stdout is None else stdout
                stderr = command_stderr if stderr is None else stderr
                self._patched_init(stdout=stdout, stderr=stderr, no_color=no_color, force_color=force_color)

            BaseCommand._patched_init = BaseCommand.__init__
            BaseCommand.__init__ = command_init_patch

            execute_from_command_line_original(argv=argv)
        finally:
            BaseCommand.__init__ = BaseCommand._patched_init

    if len(argv) > 1:
        # some arguments must be processed before django setup
        parser_args = (None,) if StrictVersion(get_main_version()) < StrictVersion('2.1') else tuple()
        parser = CommandParser(*parser_args, usage='%(prog)s subcommand [options] [args]', add_help=False)
        parser.add_argument('--settings')
        parser.add_argument('--pythonpath')
        parser.add_argument('args', nargs='*')  # catch-all
        try:
            options, args = parser.parse_known_args(argv[2:])
            handle_default_options(options)
        except CommandError:
            pass  # Ignore any option errors at this point.

        django.setup()

        return CommandExecutor(
            command_function=execute_from_command_line_with_stdout,
            command_kwargs={'argv': argv},
            name=argv[1],
            input=' '.join(argv[2:]),
            is_executed_from_command_line=True
        ).run()

    else:
        execute_from_command_line_original(argv=argv)


def call_command(command_name, stdout=None, stderr=None, *args, **options):
    from security.command import CommandExecutor

    stdout = sys.stdout if stdout is None else stdout
    stderr = sys.stderr if stderr is None else stderr

    return CommandExecutor(
        command_function=call_command_original,
        command_args=(command_name,) + tuple(args),
        command_kwargs=options,
        stdout=stdout,
        stderr=stderr,
        name=command_name,
        input=', '.join(
            list(args)+['{}={}'.format(k, v) for k, v in options.items()]
        ),
        is_executed_from_command_line=False
    ).run()



import copy

from threading import local
from attrdict import AttrDict

from uuid import uuid4

from contextlib import ContextDecorator

from django.conf import settings as django_settings

from security.config import settings


class SecurityLogger(ContextDecorator, local):

    loggers = []
    name = None
    store = True

    def __init__(self, id=None, related_objects=None, slug=None, data=None, extra_data=None):
        self.id = id or (uuid4() if self.name else None)
        self.parent = SecurityLogger.loggers[-1] if SecurityLogger.loggers else None
        self.related_objects = set(related_objects) if related_objects else set()
        self.slug = slug
        if self.parent:
            self.related_objects |= self.parent.related_objects
            if not self.slug:
                self.slug = self.parent.slug
        self.data = {}
        if data:
            self.data.update(data)
        self.parent_with_id = self._get_parent_with_id()
        self._extra_data = extra_data
        if self._extra_data is None:
            self._extra_data = self.parent.extra_data if self.parent else {}

        if self.store:
            SecurityLogger.loggers.append(self)

            if 'reversion' in django_settings.INSTALLED_APPS:
                from reversion.signals import post_revision_commit

                post_revision_commit.connect(self._post_revision_commit, weak=False)
        self.backend_logs = AttrDict()
        self.stream = None

    def _get_parent_with_id(self):
        parent = self.parent
        while parent and not parent.id:
            parent = parent.parent
        return parent

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.store:
            self.close()

    def set_slug(self, slug):
        self.slug = slug

    def add_related_objects(self, *related_objects):
        self.related_objects |= set(related_objects)

    @property
    def extra_data(self):
        return copy.deepcopy(self._extra_data)

    def update_extra_data(self, data):
        self._extra_data.update(data)

    def close(self):
        if not SecurityLogger.loggers or SecurityLogger.loggers[-1] != self:
            raise RuntimeError('Log already finished')

        SecurityLogger.loggers.pop()
        if 'reversion' in django_settings.INSTALLED_APPS:
            from reversion.signals import post_revision_commit

            post_revision_commit.disconnect(self._post_revision_commit)

    @property
    def release(self):
        return settings.RELEASE

    def _post_revision_commit(self, **kwargs):
        """
        Called as a post save of revision model of the reversion library.
        If log context manager is active input logged request, command
        log or celery task run log is joined with revision via related objects.
        """
        reversion_data = self._extra_data['reversion'] = self._extra_data.get('reversion', {
            'revisions': [],
            'total_count': 0
        })
        if reversion_data['total_count'] < settings.LOG_MAX_REVISIONS_COUNT:
            reversion_data['revisions'].append({
                'id': kwargs['revision_id'] if 'revision_id' in kwargs else kwargs['revision'].id
            })
        reversion_data['total_count'] += 1


def get_last_logger(name):
    for logger in SecurityLogger.loggers[::-1]:
        if logger.name == name:
            return logger
    return None
